
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include "airport.h"
#include "plane.h"
#include "reparse.h"
#include "dispatch.h"

//=============================================================================
//=== Static data, Common datatypes, etc. =====================================
//=============================================================================

static int dd_cur_plane = 0;        // ID of the plane we are dealing with
static plane* dd_cur_planeobj=NULL; // the plane object we are dealing with

//=============================================================================
//=== Helper Functions ========================================================
//=============================================================================

// Returns TRUE if "xx" is a letter of the alphabet
#define local_is_alpha(xx) (((xx)>='A' && (xx)<='Z')||((xx)>='a' && (xx)<='z'))

// Returns TRUE if "xx" is a numeric digit
#define local_is_digit(xx) ((xx)>='0' && (xx)<='9')

// Returns TRUE if "xx" is a "linebreak", "space", "tab",... etc.
#define local_is_space(xx) ((xx)>0 && (xx)<=32)

static char* reparse_get_string(reparse* parse,char* key)
 {
  int i; if (parse==NULL) return NULL;
  for(i=0; i<(parse->n); i++)
   {
    if (strcasecmp(parse->sub[i].key,key)==0)
     {
      if (parse->sub[i].value) return parse->sub[i].value;
      return (parse->sub[i].subvalue)->head;
     }
   }
  return NULL;
 }

static int reparse_get_int(reparse* parse,char* key)
 {
  char* ans = reparse_get_string(parse,key); if (ans==NULL) return -1;
  while(local_is_space(*ans) || *ans=='(') ans=ans+1;
  if (local_is_digit(*ans)) return atoi(ans); return (-1);
 }

static int reparse_get_twrw(reparse* parse,char* key)
 {
  // Returns 1000 for "the current runway",  returns 100..135 for runway 0..35
  // Returns 2000 for "the current taxiway", returns 201..226 for runway A..Z
  // Returns -1 for no match
  int i; if (parse==NULL) return (-1);
  for(i=0; i<(parse->n); i++)
   {
    if (strcasecmp(parse->sub[i].key,key)==0)
     {
      if (parse->sub[i].value) return (-1);
      parse=(parse->sub[i].subvalue);
      if (strcasecmp(parse->head,"runway")==0)
       {
        i=reparse_get_int(parse,"num");
        if (i<0 || i>35) return 1000; return 100+i;
       }
      if (strcasecmp(parse->head,"taxiway")==0)
       {
        key=reparse_get_string(parse,"num"); if (key==NULL) return 2000;
        if (strcasecmp(key,"alpha"   )==0) return 201;
        if (strcasecmp(key,"bravo"   )==0) return 202;
        if (strcasecmp(key,"charlie" )==0) return 203;
        if (strcasecmp(key,"delta"   )==0) return 204;
        if (strcasecmp(key,"echo"    )==0) return 205;
        if (strcasecmp(key,"foxtrot" )==0) return 206;
        if (strcasecmp(key,"golf"    )==0) return 207;
        if (strcasecmp(key,"hotel"   )==0) return 208;
        if (strcasecmp(key,"india"   )==0) return 209;
        if (strcasecmp(key,"juliet"  )==0) return 210;
        if (strcasecmp(key,"kilo"    )==0) return 211;
        if (strcasecmp(key,"lima"    )==0) return 212;
        if (strcasecmp(key,"mike"    )==0) return 213;
        if (strcasecmp(key,"november")==0) return 214;
        if (strcasecmp(key,"oscar"   )==0) return 215;
        if (strcasecmp(key,"papa"    )==0) return 216;
        if (strcasecmp(key,"quebec"  )==0) return 217;
        if (strcasecmp(key,"romeo"   )==0) return 218;
        if (strcasecmp(key,"sierra"  )==0) return 219;
        if (strcasecmp(key,"tango"   )==0) return 220;
        if (strcasecmp(key,"uniform" )==0) return 221;
        if (strcasecmp(key,"victor"  )==0) return 222;
        if (strcasecmp(key,"whiskey" )==0) return 223;
        if (strcasecmp(key,"x-ray"   )==0) return 224;
        if (strcasecmp(key,"yankee"  )==0) return 225;
        if (strcasecmp(key,"zulu"    )==0) return 226; return 2000;
       }
      return -1;
     }
   }
  return (-1);
 }

static int reparse_get_timevalue(reparse* parse,char* key)
 {
  int i;
  for(i=0; i<(parse->n); i++)
   {
    if (strcasecmp(parse->sub[i].key,key)==0)
     {
      parse=parse->sub[i].subvalue; if (parse==NULL) return -1;
      if (strcasecmp(parse->head,"timevalue")!=0) return -1;
      i=reparse_get_int(parse,"at");
      if (i>=0 && i<=24) return i-100;
      return reparse_get_int(parse,"in");
     }
   }
  return 0;
 }

//=============================================================================
//=== Dispatch Functions ======================================================
//=============================================================================

static int report_status(char* answer)
 {
  plane_state x = dd_cur_planeobj->get_state();
  if (x==TAXI)
   {
    strcpy(answer,"Roger. Currently moving towards the run-up area.");
    return 1;
   }
  if (x==RUNUP)
   {
    strcpy(answer,"Roger. Currently performing runup."); return 1;
   }
  if (x==HOLDSHORT)
   {
    strcpy(answer,"Holding short. Ready for departure."); return 1;
   }
  if (x==POSITION)
   {
    strcpy(answer,"Position and hold. Ready for departure."); return 1;
   }
  strcpy(answer,"Roger."); return 1;
 }

static int dispatch_go(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // clear to runway five       ==> (go :on (runway  :num 5       ) :agent you)
  // clear to taxi to five      ==> (go :on (runway  :num 5       ) :agent you)
  // cleared direct to runway 5 ==> (go :on (runway  :num 5       ) :agent you)
  // cleared to runway five     ==> (go :on (runway  :num 5       ) :agent you)
  // continue on tango          ==> (go :on (taxiway :num tango   ) :agent you)
  // go november                ==> (go :on (taxiway :num november) :agent you)
  // go straight down tango     ==> (go :on (taxiway :num tango   ) :agent you)
  // exit on taxiway tango      ==> (go :on (taxiway :num tango   ) :agent you)
  // straight ahead on zulu     ==> (go :on (taxiway :num zulu    ) :agent you)
  // straight on to runway 3    ==> (go :on (runway  :num 3       ) :agent you)
  // straight on to zulu        ==> (go :on (taxiway :num zulu    ) :agent you)
  // taxi ahead on zulu         ==> (go :on (taxiway :num zulu    ) :agent you)
  // taxi quebec                ==> (go :on (taxiway :num quebec  ) :agent you)
  // then runway three          ==> (go :on (runway  :num 3       ) :agent you)
  // then zulu                  ==> (go :on (taxiway :num zulu    ) :agent you)
  // taxi to the runway         ==> (go :on (runway)                :agent you)
  //
  // continue on zulu until runway two
  // ==> (go :on (taxiway :num zulu) :until (runway :num 2) :agent you)
  //
  // taxi to runway three via zulu
  // ==> (go :on (runway :num 3) :via (taxiway :num zulu) :agent you)
  //
  // taxi via taxiway tango
  // ==> (go :via (taxiway :num tango) :agent you)
  //
  // ======= Sentences ==========
  //
  int my_on  = reparse_get_twrw(parse,"on");
  int my_via = reparse_get_twrw(parse,"via");
  if (my_on>=0 && my_via>=0)
   {
    if ((my_on>=100 && my_on<=135) && (my_via>=201 && my_via<=226))
     {
      int x1=airport_runway_to_point(my_on-100),x2=(my_via-200);
      if (x1<0) goto bad_runway;
      if (airport_taxiway_to_points(x2)==NULL) goto bad_taxiway;
      dd_cur_planeobj->set_dest(x1);
      dd_cur_planeobj->via(x2); goto good;
     }
    else
     {
      int x1=airport_runway_to_point(my_on-100);
      if (x1<0) goto bad_runway; dd_cur_planeobj->set_dest(x1); goto good;
     }
   }
  if (my_on>=0)
   {
    if (my_on>=100 && my_on<=135)
     {
      int x=airport_runway_to_point(my_on-100);
      if (x>=0) { dd_cur_planeobj->set_dest(x); goto good; } goto bad_runway;
     }
    else
     {
      int x=(my_on-200);
      if (airport_taxiway_to_points(x)!=NULL)
         { dd_cur_planeobj->via(x); goto good; }
      goto bad_taxiway;
     }
   }
  return 0;
  bad_runway:
  strcpy(answer,"I do not think there is such a runway."); return 1;
  bad_taxiway:
  strcpy(answer,"I do not think there is such a taxiway."); return 1;
  good: report_status(answer); return 1;
 }

//=============================================================================

static int dispatch_position(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // position and hold           ==> (position :agent you) + (hold :agent you)
  // taxi into position and hold ==> (position :agent you) + (hold :agent you)
  // taxi to position and hold   ==> (position :agent you) + (hold :agent you)
  // ======= Sentences ==========
  plane_state x = dd_cur_planeobj->get_state();
  if (x==PARK)
   {
    strcpy(answer,"Please advice which runway to go to."); return 1;
   }
  if (x==TAXI)
   {
    strcpy(answer,"Still moving towards the run-up area."); return 1;
   }
  if (x!=HOLDSHORT)
   {
    strcpy(answer,"Still performing runup."); return 1;
   }
  dd_cur_planeobj->speed=1;
  report_status(answer); return 1;
 }

//=============================================================================

static int dispatch_setfreq(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // contact three          ==> setfreq :goal (speaker :freq 3                )
  // contact departure      ==> setfreq :goal (speaker :name departure        )
  // contact departure at 3 ==> setfreq :goal (speaker :name departure :freq 3)
  // contact ground         ==> setfreq :goal (speaker :name ground           )
  // contact ground on five ==> setfreq :goal (speaker :name ground :freq 5   )
  // contact ramp           ==> setfreq :goal (speaker :name ramp             )
  // contact ramp on ten    ==> setfreq :goal (speaker :name ramp :freq 10    )
  // contact tower          ==> setfreq :goal (speaker :name tower            )
  // contact tower on eight ==> setfreq :goal (speaker :name tower :freq 8    )
  // monitor four           ==> setfreq :goal (speaker :freq 4                )
  // monitor ground         ==> setfreq :goal (speaker :name ground           )
  // monitor ramp           ==> setfreq :goal (speaker :name ramp             )
  // monitor tower          ==> setfreq :goal (speaker :name tower            )
  // remain on ten          ==> setfreq :goal (speaker :freq 10               )
  // ======= Sentences ==========
  int i; char* x;
  for(i=0; i<(parse->n); i++)
   {
    if (strcasecmp(parse->sub[i].key,"goal")==0)
     {
      parse=parse->sub[i].subvalue; if (parse==NULL) return 0;
      if (strcasecmp(parse->head,"speaker")!=0) return 0;
      i=reparse_get_int(parse,"freq");
      x=reparse_get_string(parse,"name");
      if (x==NULL) { if (i>=0) goto good; return 0; }
      if (strcasecmp(x,"ground")==0)
       {
        int j=airport_get_ground(); if (i<0 || i==j) goto good;
        sprintf(answer,"Please confirm. "
          "The frequency you gave is not the ground frequency (%d) I know.",j);
        return 1;
       }
      if (strcasecmp(x,"tower")==0)
       {
        int j=airport_get_tower(); if (i<0 || i==j) goto good;
        sprintf(answer,"Please confirm. "
          "The frequency you gave is not the tower frequency (%d) I know.",j);
        return 1;
       }
      sprintf(answer,"Please give me the exact frequency."); return 1;
     }
   }
  return 0;
  good: dd_cur_planeobj->set_freq(i);
  sprintf(answer,"Roger. Frequency switched to %d.",i); return 1;
 }

//=============================================================================

static int dispatch_enable_takeoff(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // clear for takeoff          ==> (enable-takeoff :agent you)
  // clear to takeoff runway 29 ==> (takeoff :on (runway :num 29) :agent you)
  // ======= Sentences ==========
  plane_state x = dd_cur_planeobj->get_state();
  if (x==PARK)
   {
    strcpy(answer,"Please advice which runway to go to."); return 1;
   }
  if (x==TAXI)
   {
    strcpy(answer,"Still moving towards the run-up area."); return 1;
   }
  if (x==RUNUP)
   {
    strcpy(answer,"Still performing runup."); return 1;
   }
  if (x==HOLDSHORT)
   {
    dd_cur_planeobj->speed=1;
    strcpy(answer,"Roger. Moving towards the runway."); return 1;
   }
  if (x==POSITION)
   {
    dd_cur_planeobj->set_state(TAKINGOFF);
    dd_cur_planeobj->speed=5;
   }
  report_status(answer); return 1;
 }

//=============================================================================

static int helper_find_plane(char* goal);

static int dispatch_ahead(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // move ahead before the boeing"
  //  ==> (move :agent you) + (ahead :goal (boeing) :agent you)
  //
  // right in front of the continental"
  //  ==> (move :agent you) + (ahead :goal (continental) :agent you)
  // ======= Sentences ==========
  //
  int plane; char* goal=reparse_get_string(parse,"goal");
  if (goal==0) return 0;
  plane=helper_find_plane(goal);
  if (plane<=0)
   {
    strcpy(answer
      ,"Sorry, I can not find an airplane matching that description.");
    return 1;
   }
  dd_cur_planeobj->ahead(plane);
  report_status(answer); return 1;
 }

static int dispatch_behind(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // After the continental        ==> (behind :goal (continental) :agent you)
  // fall in behind the alaska    ==> (behind :goal (alaska     ) :agent you)
  // follow in behind that alaska ==> (behind :goal (alaska     ) :agent you)
  // follow that boeing           ==> (behind :goal (boeing     ) :agent you)
  // get behind the continental   ==> (behind :goal (continental) :agent you)
  // give way to boeing           ==> (behind :goal (boeing     ) :agent you)
  // hold for the continental     ==> (behind :goal (continental) :agent you)
  // give way to the boeing       ==> (behind :goal (boeing     ) :agent you)
  // you are following a alaska   ==> (behind :goal (alaska     ) :agent you)
  //
  // let the boeing turn in front of you ==> (behind :goal (boeing) :agent you)
  // the boeing is going in front of you ==> (behind :goal (boeing) :agent you)
  //
  // follow a continental that is behind a boeing
  //   ==> (behind :goal (continental :behind boeing) :agent you)
  //
  // follow that continental directly ahead of you
  //   ==> (behind :goal (continental :ahead you) :agent you)
  //
  // follow the boeing from your left
  //   ==> (behind :goal (boeing :left you) :agent you)
  //
  // ======= Sentences ==========
  //
  int plane; char* goal=reparse_get_string(parse,"goal");
  if (goal==0) return 0;
  plane=helper_find_plane(goal);
  if (plane<=0)
   {
    strcpy(answer
      ,"Sorry, I can not find an airplane matching that description.");
    return 1;
   }
  dd_cur_planeobj->behind(plane);
  report_status(answer); return 1;
 }

//=============================================================================

static int dispatch_altimeter(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // Altimeter 6 ==> (altimeter :value (6) :agent you)
  // ======= Sentences ==========
  //
  int alt=reparse_get_int(parse,"value");
  if (alt>=0)
   {
    dd_cur_planeobj->set_altimeter(alt);
    sprintf(answer,"Roger. Altimeter set to %d.",alt); return 1;
   }
  return 0;
 }

//=============================================================================

static int dispatch_undo(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // "cancel that"     ==> (undo :agent you)
  // "change of plans" ==> (undo :agent you)
  // ======= Sentences ==========
  //
  switch (dd_cur_planeobj->get_newly_changed())
   {
    case LAST_CHANGE_IS_FREQ:
      dd_cur_planeobj->swap_freq();
      sprintf(answer,"Roger. Changing back to the previous frequency of %d."
        ,dd_cur_planeobj->get_freq());
      return 1;

    case LAST_CHANGE_IS_SPEED:
      dd_cur_planeobj->swap_speed();
      sprintf(answer,"Roger. Changing back to the previous target speed of %d."
        ,dd_cur_planeobj->get_target_speed());
      return 1;

    case LAST_CHANGE_IS_DEST:
      dd_cur_planeobj->speed=0;
      strcpy(answer,"Roger. Awaiting a new destination.");
      return 1;

    default: report_status(answer); return 1;
   }
 }

//=============================================================================

static int dispatch_setspeed(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // change speed to six        ==> (setspeed :goal (6) :agent you)
  // change speed to six knots  ==> (setspeed :goal (6) :agent you)
  // ======= Sentences ==========
  int velocity = reparse_get_int(parse,"goal");
  if (velocity<0) return 0;
  if (dd_cur_planeobj->is_speed_locked())
   {
    strcpy(answer,
      "I can not change my velocity, since that setting is locked.");
    return 1;
   }
  dd_cur_planeobj->set_target_speed(velocity);
  sprintf(answer,"Roger. Speed changed to %d knots.",velocity);
  return 1;
 }

//=============================================================================

static int dispatch_checkin(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // checkin at five o^clock ==> (checkin :time (timevalue :at 5) :agent you)
  // checkin in six  minutes ==> (checkin :time (timevalue :in 6) :agent you)
  // ======= Sentences ==========
  int time = reparse_get_timevalue(parse,"time"); if (time==(-1)) return 0;
  if (time<0)
   {
    sprintf(answer,"Roger. I will checkin at %d o'clock.",time+100);
   }
  else
   {
    sprintf(answer,"Roger. I will checkin in %d minutes.",time);
   }
  return 1;
 }

//=============================================================================

static int dispatch_land(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // clear to land                  ==> (land :agent you)
  // clear to land runway five      ==> (land :on   (runway :num 5) :agent you)
  // clear to land runway five left ==> (land :left (runway :num 5) :agent you)
  // clear for landing runway 5     ==> (land :on   (runway :num 5) :agent you)
  // ======= Sentences ==========
  plane_state x = dd_cur_planeobj->get_state();
  if (x!=OFFSCREEN)
   {
    strcpy(answer,"I am not currently in the air."); return 1;
   }
  int my_on    = reparse_get_twrw(parse,"on");
  int my_left  = reparse_get_twrw(parse,"left");
  int my_right = reparse_get_twrw(parse,"right");
  int r=(-1);
  if (r<0) if (my_on   >= 100 && my_on   <= 135) r=my_on   -100;
  if (r<0) if (my_left >= 100 && my_left <= 135) r=my_left -100;
  if (r<0) if (my_right>= 100 && my_right<= 135) r=my_right-100;
  if (r>=0 && r<=35) if (airport_runway_to_point(r)<0) r=(-1);
  if (r<0)
   {
    strcpy(answer,"Please specify a runway on this airport."); return 1;
   }
  dd_cur_planeobj->land(r);
  dd_cur_planeobj->speed=1;
  sprintf(answer,"Roger. Preparing to land on runway %d.",r);
  return 1;
 }

//=============================================================================

static int dispatch_climb_maintain(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // climb and maintain three      ==> (climb-maintain :alt (3) :agent you)
  // climb and maintain three feet ==> (climb-maintain :alt (3) :agent you)
  // ======= Sentences ==========
  strcpy(answer,"Roger. Climbing and maintaining altitude."); return 1;
  // No work needed here, since our simulator does not yet
  // display in-air aircraft motions, so the movement is invisible.
 }

//=============================================================================

static int dispatch_descend_maintain(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // descend and maintain three      ==> (descend-maintain :alt (3) :agent you)
  // descend and maintain three feet ==> (descend-maintain :alt (3) :agent you)
  // ======= Sentences ==========
  strcpy(answer,"Roger. Descending and maintaining altitude."); return 1;
  // No work needed here, since our simulator does not yet
  // display in-air aircraft motions, so the movement is invisible.
 }

//=============================================================================

static int dispatch_leave(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // clear the runway  ==> (leave :goal (runway ) :agent you)
  // clear the taxiway ==> (leave :goal (taxiway) :agent you)
  // exit the ramp     ==> (leave :goal ramp      :agent you)
  // ======= Sentences ==========
  report_status(answer); return 1;
  // No work needed here, since each simulated plane will already
  // plot the shortest path to get to their final destination,
  // and move ahead as soon as they can. So no need to "rush" them.
 }

//=============================================================================

static int dispatch_move(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // move ahead              ==> (move :agent you)
  // move as soon as you can ==> (move :agent you)
  // taxi ahead              ==> (move :agent you)
  // ======= Sentences ==========
  report_status(answer); return 1;
  //
  // No work needed here, since our simulated airplanes
  // already move as much as they can towards their goal
  // at all time. There's no need to tell them to "move ahead".
  //
  // On the other hand, if a plane is waiting for a clearance,
  // a simple "move ahead" is not sufficient to convince a plane
  // that it can now use the runway (without contending with anyone else).
  // So it still prefers to wait for a "clear for takeoff", or something
  // to that effect.
 }

//=============================================================================

static int dispatch_cross(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // cross two at five knots ==> (cross :fix 2 :at-speed    5 :agent you)
  // cross two at three feet ==> (cross :fix 2 :at-altitude 3 :agent you)
  //
  // cross two               ==> (cross :road (runway :num 2) :agent you)
  // cross runway two        ==> (cross :road (runway :num 2) :agent you)
  //
  // cross runway three at zulu
  //   ==> (cross :road (runway :num 3) :at (taxiway :num zulu) :agent you)
  //
  // ======= Sentences ==========
  report_status(answer); return 1;
  // No work needed here, since our simulator does not yet
  // display in-air aircraft motions, so the movement is invisible.
 }

//=============================================================================

static int dispatch_expect(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // expect boeing on tango
  //  ==> (expect :plane (boeing) :on (taxiway :num tango) :agent you)
  //
  // expect traffic on november
  //  ==> (expect :on (taxiway :num november) :agent you)
  //
  // you are going to encounter a alaska
  //  ==> (expect :plane (alaska) :agent you)
  //
  // you are going to see an alaska behind a boeing
  //  ==> (expect :plane (alaska :behind boeing) :agent you)
  //
  // you are going to see an alaska
  //  ==> (expect :plane (alaska) :agent you)
  // ======= Sentences ==========
  report_status(answer); return 1;
  //
  // This message is for the plane to watch out for other crafts.
  //
  // Since our computer-controlled planes already stop themselves
  // whenever they get too close to another aircraft, this function
  // is a no-op.
 }

//=============================================================================

static int dispatch_hold_heading(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // hold five         ==> (hold-heading :heading (heading :to 5) :agent you)
  // hold heading five ==> (hold-heading :heading (heading :to 5) :agent you)
  // ======= Sentences ==========
  report_status(answer); return 1;
  // No work needed here, since our simulator does not yet
  // display in-air aircraft motions, so the heading is invisible.
 }

//=============================================================================

static int dispatch_hold_short(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // hold short ==> (hold-short :agent you)
  //
  // hold short of runway three on tango
  //  ==> (hold-short :of (runway :num 3) :on (taxiway :num tango) :agent you)
  //
  // hold short of taxiway zulu
  //  ==> (hold-short :of (taxiway :num zulu) :agent you)
  //
  // hold short of zulu for spacing
  //  ==> (hold-short :of (taxiway :num zulu) :agent you)
  //
  // hold short of zulu for the continental
  //  ==> (hold-short :of (taxiway :num zulu) :for (continental) :agent you)
  //
  // ======= Sentences ==========
  report_status(answer); return 1;
  // No work needed here, since planes will always hold short,
  // until they're given permission to move into runway position.
 }

//=============================================================================

static int dispatch_hold(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // position and hold           ==> (position :agent you) + (hold :agent you)
  // taxi into position and hold ==> (position :agent you) + (hold :agent you)
  // taxi to position and hold   ==> (position :agent you) + (hold :agent you)
  // ======= Sentences ==========
  report_status(answer); return 1;
  // No work needed here, since planes will always hold,
  // until they're given permission.
 }

//=============================================================================

static int dispatch_maintain(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // maintain four feet ==> (maintain :alt (4) :agent you)
  //
  // maintain heading of four ==> (maintain :heading 4 :agent you)
  //
  // maintain five feet at departure
  // on departure maintain five
  // on departure maintain five feet
  //   ==> (maintain :alt (5) :when on-departure :agent you)
  // ======= Sentences ==========
  report_status(answer); return 1;
  // No work needed here, since our simulator does not yet
  // display in-air aircraft motions, so the heading is invisible.
 }

//=============================================================================

static int dispatch_turn(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // turn heading three
  // turn three
  // turn to heading three"
  //   ==> (turn :heading (heading :to 3) :agent you)
  //
  // turn left five degrees"
  //   ==> (turn :heading (heading :left 5) :agent you)
  //
  // turn left five degrees to heading ten"
  //   ==> (turn :heading (heading :left 5 :to 10) :agent you)
  //
  // turn left to heading three"
  //   ==> (turn :heading (heading :left unknown :to 3) :agent you)
  // ======= Sentences ==========
  strcpy(answer,"Roger."); return 1;
  // No work needed here, since our simulator does not yet
  // display in-air aircraft motions, so the heading is invisible.
 }

//=============================================================================

static int dispatch_set_priority(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // you are number five
  //   ==> (set-priority :new (priority :num 5 :event unknown))
  //
  // you are number five behind a cessna
  //   ==> (set-priority :new (priority :num 5 :event unknown :behind cessna))
  //
  // you are number five for landing
  //   ==> (set-priority :new (priority :num 5 :event landing))
  //
  // you are number five for landing behind a boeing
  //   ==> (set-priority :new (priority :num 5 :event landing :behind boeing))
  //
  // you are number five for takeoff
  //   ==> (set-priority :new (priority :num 5 :event takeoff))
  //
  // you are number five for takeoff behind a cessna
  //   ==> (set-priority :new (priority :num 5 :event takeoff :behind cessna))
  //
  // you are number five to go
  //   ==> (set-priority :new (priority :num 5 :event unknown))
  // ======= Sentences ==========
  strcpy(answer,"Roger."); return 1;
 }

//=============================================================================

static int dispatch_lock_altitude(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // expect three feet five minutes after departure
  //   ==> (lock-altitude :alt 3 :time 5 :agent you)
  // ======= Sentences ==========
  strcpy(answer,"Roger."); return 1;
 }

//=============================================================================

static int dispatch_intercept(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // intercept five ==> (intercept :patient 5 :agent you)
  // ======= Sentences ==========
  strcpy(answer,"Roger."); return 1;
 }

//=============================================================================

static int dispatch_nop(reparse* parse,char* answer)
 {
  // ======= Sentences ==========
  // maintain this frequency ==> (nop :agent you)
  // remain this frequency   ==> (nop :agent you)
  // ======= Sentences ==========
  strcpy(answer,"Roger."); return 1;
 }

static int helper_find_plane(char* goal)
 {
  // TODO: Need to search the plane database
  // (taking into account if the user said things like
  // "the alaska that is behind a Cessna...")
  return 1;
 }

//=============================================================================
//=== Dispatch Table ==========================================================
//=============================================================================

typedef int (*DispatchFn) (reparse* parse,char* answer);

typedef struct
 {
  char* verb;
  DispatchFn fnptr;
 }
dispatch_fn_entry;

static dispatch_fn_entry local_table[]
 = { { "altimeter"        , dispatch_altimeter        }
   , { "ahead"            , dispatch_ahead            }
   , { "behind"           , dispatch_behind           }
   , { "undo"             , dispatch_undo             }
   , { "setspeed"         , dispatch_setspeed         }
   , { "checkin"          , dispatch_checkin          }
   , { "enable-takeoff"   , dispatch_enable_takeoff   }
   , { "takeoff"          , dispatch_enable_takeoff   }
   , { "leave"            , dispatch_leave            }
   , { "land"             , dispatch_land             }
   , { "go"               , dispatch_go               }
   , { "climb-maintain"   , dispatch_climb_maintain   }
   , { "descend-maintain" , dispatch_descend_maintain }
   , { "setfreq"          , dispatch_setfreq          }
   , { "move"             , dispatch_move             }
   , { "cross"            , dispatch_cross            }
   , { "expect"           , dispatch_expect           }
   , { "hold-heading"     , dispatch_hold_heading     }
   , { "hold-short"       , dispatch_hold_short       }
   , { "hold"             , dispatch_hold             }
   , { "position"         , dispatch_position         }
   , { "maintain"         , dispatch_maintain         }
   , { "turn"             , dispatch_turn             }
   , { "set-priority"     , dispatch_set_priority     }
   , { "lock-altitude"    , dispatch_lock_altitude    }
   , { "intercept"        , dispatch_intercept        }
   , { "nop"              , dispatch_nop              }
   , { ""                 , dispatch_nop              } };

//=============================================================================
//=== Dispatch Table ==========================================================
//=============================================================================

void process_line(int ac,char* parse,char* answer)
 {
  int i;
  plane* temp=plane::getplane(ac);
  if (temp==NULL)
   {
    strcpy(answer,"Please clarify which plane you are talking to."); return;
   }
  dd_cur_planeobj=temp; dd_cur_plane=ac;
  reparse* reparse=reparse_parse(parse);
  if (reparse==NULL) goto bad;
  for(i=0;*(local_table[i].verb);i=i+1)
   if (strcasecmp(local_table[i].verb,reparse->head)==0)
      { i=local_table[i].fnptr(reparse,answer); if (i==0) break; return; }
  bad: strcpy(answer,"I can not understand the sentence. Please repeat.");
 }

//=============================================================================
