;;; -*- mode: lisp; base: 10; syntax: common-lisp -*-
;;;



(defvar *rule-sets* nil)
(defvar *valid-rule-types* '(dictionary cfg cfg-table gpsg gpsg-table sentences))

(defun rule-set (name &optional type)
  (let ((rule-set (cdr (assoc name *rule-sets*))))
    (if (or (not type) (eq type (second rule-set)))
	rule-set
	nil)))

(defun add-rule-set (name type)
  (cond ((member type *valid-rule-types*)
	 (when (rule-set name)
	   (remove-rule-set name))
	 (push (cons name (list name type nil)) *rule-sets*)
	 (msg t "~&rule set ~s added." name))
	(t (format t "~&illegal type of rule set.")))
  (values))

(defun remove-rule-set (name)
  (setq *rule-sets* (remove name *rule-sets* :key #'car))
  (msg t "~&rule set ~s removed." name)
  (values))

(defun sort-rules (name)
  (if (rule-set name)
      (setf (third (rule-set name)) (sort (copy-list (third (rule-set name))) #'s<))
      (format t "~&nonexistant rule set."))
  (values))

(let ((display-table nil))
  (defun list-rules (name &optional (template '*))
    (let ((rs (rule-set name)))
      (cond ((member (second rs) '(cfg-table gpsg-table))
	     (format t "~&~s is not a listable rule set." name))
	    (rs
	     (format t "~&~s rule set ~s:" (second rs) name)
	     (let ((to-display (sort (copy-list (filt (third rs) template)) #'s<)))
	       (setq display-table (list rs (length to-display) to-display))
	       (do ((to-do to-display (cdr to-do))
		    (n 1 (1+ n)))
		   ((null to-do))
		 (cond ((symbolp (car to-do)) (format t "~&~d.~5t~s" n (car to-do)))
		       ((stringp (car to-do)) (format t "~&~d.~6t~a" n (car to-do)))
		       (t (format t "~&~d.~5t~{ ~s~}" n (car to-do)))))))
	    (t (format t "~&no such rule set."))))
    (values))

  (defun del (&rest nums)
    (if (or (null display-table) (= 0 (second display-table)))
	(format t "~&use list-rules to specify a rule set.")
	(mapc #'(lambda (i) (if (and (integerp i) (> i 0) (<= i (second display-table)))
				(del-rule (first (first display-table))
					  (nth (1- i) (third display-table)))
				(format t "~&a reference must be between 1 and ~d."
					(second display-table))))
	      nums))
    (values))

  (defun rpl (&rest nums-rules)
    (if (or (null display-table) (= 0 (second display-table)))
	(format t "~&use list-rules to specify a rule set.")
	(do ((to-do (copy-list nums-rules) (cddr to-do)))
	    ((or (null to-do) (null (cdr to-do))))
	  (if (and (integerp (car to-do)) (> (car to-do) 0)
		   (<= (car to-do) (second display-table)))
	      (progn (del-rule (first (first display-table))
			       (nth (1- (car to-do)) (third display-table)))
		     (add-rule (first (first display-table))
			       (second to-do)))
	      (format t "~&a reference must be between 1 and ~d."
		      (second display-table)))))
    (values))

  (defun add (&rest rules)
    (if (null display-table)
	(format t "~&use list-rules to specify a rule set.")
	(add-rule-list (first (first display-table)) (copy-list rules)))
    (values)))

(defun list-all ()
  (format t "~%~%")
  (let ((n 0))
    (dolist (rule-set *rule-sets*)
      (incf n)
      (format t "~d.  ~5t~s~20t~s~40t~d~%" n (second rule-set)
	      (third rule-set) (length (fourth rule-set)))))
  (values))

(defun add-rule (name rule)
  (let ((rs (rule-set name)))
    (cond ((null rs) (format t "~%nonexistant rule set.~%"))
	  ((illegal-rule-p (second rs) rule)
	   (format t "~%illegal format for ~s rule: ~s" (second rs) rule))
	  (t (setf (third rs) (adjoin rule (third rs) :test #'equal))
	     (msg t "~&adding rule ~s to rule set ~s." rule name))))
  (values))

(defun set-rule-list (name rule-list)
  (if (rule-set name)
      (setf (third (rule-set name)) rule-list)
      (format t "~%nonexistant rule set.~%"))
  nil)

(defun del-rule (name rule)
  (let ((rs (rule-set name)))
    (cond ((null rs) (format t "~%nonexistant rule set.~%"))
	  ((illegal-rule-p (second rs) rule)
	   (format t "~%illegal format for ~s rule." (second rs)))
	  (t (setf (third rs) (remove rule (third rs) :test #'equal))
	     (msg t "~&deleting rule ~s from rule set ~s." rule name))))
  (values))

(defun copy-rule-set (name-from name-to)
  (let ((rs1 (rule-set name-from))
	(rs2 (rule-set name-to)))
    (cond ((null rs1) (format t "~%~s does not exist.~%" name-from))
	  (rs2 (msg t "~&~s already exists." name-to))
	  (t (add-rule-set name-to (second rs1))
	     (direct-add-rule-list name-to (third rs1)))))
  (values))

(defun merge-rule-sets (name-from name-to)
  (let ((rs1 (rule-set name-from))
	(rs2 (rule-set name-to)))
    (cond ((null rs1) (format t "~&~s does not exist." name-from))
	  ((null rs2) (format t "~&~s does not exist." name-to))
	  ((not (eq (second rs1) (second rs2)))
	   (format t "~&~s and ~s are of different types." name-from name-to))
	  (t (direct-add-rule-list name-to (third rs1)))))
  (values))

(defun illegal-rule-p (rule-type rule)
  (cond ((eq rule-type 'sentences)
	 (and (not (stringp rule))
	      (not (and (listp rule) (every #'symbolp rule)))))
	((eq rule-type 'gpsg-table)
	 (or (not (listp rule))
	     (not (= (length rule) 6))))
	((eq rule-type 'cfg-table)
	 (or (not (listp rule))
	     (not (= (length rule) 5))))
	((not (listp rule)) t)
	((eq rule-type 'dictionary)
	 (not (symbolp (car rule))))
	((not (eq (second rule) '==>)) t)
	((eq rule-type 'cfg)
	 (not (every #'symbolp rule)))
	((eq rule-type 'gpsg)
	 (or (not (gpsg-exp (first rule)))
	     (not (every #'gpsg-exp (cddr rule)))))
	(t t)))

(defun gpsg-exp (exp)
  (cond ((symbolp exp) t)
	((not (listp exp)) nil)
	((symbolp (car exp)) t)
	(t nil)))

(defun add-rule-list (name rule-list)
  (if (rule-set name)
      (mapc #'(lambda (rule) (add-rule name rule)) rule-list)
      (format t "~%nonexistant rule set.~%"))
  nil)

(defun del-rule-list (name rule-list)
  (if (rule-set name)
      (mapc #'(lambda (rule) (del-rule name rule)) rule-list)
      (format t "~%nonexistant rule set.~%"))
  nil)

(defun direct-add-rule-list (name rule-list)
  (if (rule-set name)
      (setf (third (rule-set name))
	    (union (third (rule-set name)) rule-list :test #'equal))
      (format t "~%nonexistant rule set.~%"))
  nil)

(defun direct-del-rule-list (name rule-list)
  (let ((g (third (rule-set name))))
    (if g (setf g (apply #'nconc (mapcar #'(lambda (i) (if (member i rule-list :test #'equal)
							   nil i))
					 g))))
    nil))

(defun filt (lst &optional (template '*))
  (remove-if-not #'(lambda (item) (filt-1 item template)) lst))

(defun filt-1 (item template)
  (cond ((eq template '*) t)
	((eq template '**) t)
	((eq item template) t)
	((not (listp item)) nil)
	((eq (car item) template) t)
	((symbolp template) nil)
	((eq (car template) '**) t)
	((and (filt-1 (car item) (car template))
	      (filt-1 (cdr item) (cdr template))) t)
	(t nil)))

(defun save-rules (path-name &rest particulars)
  (let ((to-save (if particulars
		     (mapcar #'(lambda (name) (rule-set name)) particulars)
		     (mapcar #'cdr *rule-sets*))))
    (with-open-file (out path-name :direction :output :if-exists :new-version)
      (format out ";;; -*- mode: lisp; package: gpsg; syntax: common-lisp; base: 10 -*- ~%~%")
      (format out ";;; grammar data file~%")
      (mapc #'(lambda (rs)
		(cond ((member (second rs) '(cfg gpsg dictionary rtn sentences))
		       (format out "~%; name: ~s  type: ~s~%~%" (first rs) (second rs))
		       (format out "(add-rule-set '~s '~s)~%" (first rs) (second rs))
		       (format out "(add-rule-list '~s~%  '(~{~s~%    ~}))~%"
			       (first rs) (third rs))
		       (msg t "~&~s saved." (first rs)))
		      (t (format t "~&impossible to write ~s to lisp file." (first rs)))))
	    to-save)))
  (values))

;; *******************************************
;;
;;  dictionary
;;
;; *******************************************

(defun dictionary (dict word)
  ;; the dictionary must return a symbol or list of symbols that represent the word's
  ;; possible categories.
  (let ((defs nil))
    (dolist (d dict)
	    (if (equalp (car d) word) (setq defs (append (cdr d) defs))))
    defs))

;; *******************************************
;;
;;  earley's algorithm
;;
;; *******************************************

(defun create-cfg-table (table-name cfg-name root k)
  (let ((cfg (rule-set cfg-name 'cfg))
	(table (rule-set table-name)))
    (cond ((null cfg) (format t "~%there is no cfg named ~s.~%" cfg-name))
	  (table (format t "~%illegal name for table.~%"))
	  ((not (or (integerp k) (= k 1) (= k 0)))
	   (format t "~%k must be either 0 or 1.~%"))
	  (t (add-rule-set table-name 'cfg-table)
	     (let* ((productions (cons `(*d0* ==> ,root $) (third cfg)))
		    (production-table (gensym "production-table")))
	       (dolist (p productions)
		 (dolist (d p)
		   (setf (get d production-table) nil)))
	       (let ((non-terminals nil)
		     (cpp (make-array (1+ (length (third cfg)))))
		     (dp (make-array (1+ (length (third cfg)))))
		     (first (if (= k 1) (make-array (1+ (length (third cfg))))))
		     (p 0))
		 (dolist (production productions)
		   ;; the production table pairs off non-terminals n with rule numbers q s.t.
		   ;;   dq = n.
		   (push p (get (car production) production-table))
		   (setq non-terminals (adjoin (car production) non-terminals))
		   (setf (aref cpp p) (cddr production))
		   (setf (aref dp p) (car production))
		   (incf p))
		 (when (= k 1)
		   (make-first-table first dp cpp (length (third cfg)) non-terminals))
		 (add-rule table-name (list production-table non-terminals cpp dp first)))))))
  nil)

(defun create-gpsg-table (table-name gpsg-name root k)
  (let ((gpsg (rule-set gpsg-name 'gpsg)))
    (cond ((null gpsg) (format t "~%there is no gpsg named ~s.~%" gpsg-name))
	  ((not (or (integerp k) (= k 1) (= k 0)))
	   (format t "~%k must be either 0 or 1.~%"))
	  (t (add-rule-set table-name 'gpsg-table)
	     (let* ((gp (get-gpsg (expand-gpsg (third gpsg))))
		    (production-table (gensym "production-table"))
		    (productions (cons `(*d0* ==> ,root $) (car gp))))
	       (dolist (p productions)
		 (dolist (d p)
		   (setf (get d production-table) nil)))
	       (let ((non-terminals nil)
		     (cpp (make-array (1+ (length (car gp)))))
		     (dp (make-array (1+ (length (car gp)))))
		     (gpsg-table (make-array (1+ (length (car gp)))))
		     (first (if (= k 1) (make-array (1+ (length (car gp))))))
		     (p 0))
		 (dolist (production productions)
		   (setf (aref gpsg-table p) (cdr (assoc production (second gp))))
		   ;; the production table pairs off non-terminals n with rule numbers q s.t.
		   ;;   dq = n.
		   (push p (get (car production) production-table))
		   (setq non-terminals (adjoin (car production) non-terminals))
		   (setf (aref cpp p) (cddr production))
		   (setf (aref dp p) (car production))
		   (incf p))
		 (when (= k 1)
		   (make-first-table first dp cpp (length (car gp)) non-terminals))
		 (add-rule table-name (list production-table non-terminals cpp dp first
					    gpsg-table))))))
    nil))

(defun make-first-table (first dp cpp size non-terminals)
  ;;
  ;; this computes the table first, which matches symbols with all the terminals
  ;; that could be leftmost in an arbitrary expansion of them.
  ;;
  (dotimes (p (1+ size))
    (setf (get (aref dp p) 'firsts) nil))
  ;; first compute simple emptys.
  (dotimes (p (1+ size))
    (if (null (aref cpp p))
	(setf (get (aref dp p) 'firsts)
	      (adjoin '*empty* (get (aref dp p) 'firsts)))))
  ;; now do complicated stuff until no members are added to first(dp).
  (do ((i 0 (1+ i))
       (number-added 1 number-added))
      ((zerop number-added))
    (unless (zerop i) (msg t "~%round ~d. ~d added." i number-added))
    (setq number-added 0)
    (dotimes (p (1+ size))
      (let ((d (aref dp p))
	    (cp (aref cpp p))
	    (first-dp (get (aref dp p) 'firsts)))
	(when cp
	  (do ((cpq cp (cdr cpq)))
	      ((or (null cpq) ;; end of expansion
		   (not (member (car cpq) non-terminals)) ;; cpq is a terminal
		   (not (member '*empty* (get (car cpq) 'firsts)))); cpq not empty
	       (cond ((null cpq) ;; expansion can be empty.
		      (cond ((not (member '*empty* first-dp))
			     (setf (get d 'firsts) (push '*empty* first-dp))
			     (incf number-added))))
		     ((not (member (car cpq) non-terminals)) ;; cpq is a terminal
		      (cond ((not (member (car cpq) first-dp))
			     (setf (get d 'firsts) (push (car cpq) first-dp))
			     (incf number-added))))
		     (t ;; cpq not empty
		      (let* ((first-cpq (get (car cpq) 'firsts))
			     (c (union first-dp first-cpq)))
			(unless (= (length first-dp) (length c))
			  (setq first-dp c)
			  (setf (get d 'firsts) first-dp)
			  (incf number-added))))))
	    ;; cpq is a non-terminal which may be empty.
	    (let* ((first-cpq (get (car cpq) 'firsts))
		   (c (union first-dp first-cpq)))
	      (unless (= (length first-dp) (length c)) ;; nothing has been added.
		(setq first-dp c) ;; something has been added.
		(setf (get d 'firsts) first-dp)
		(incf number-added))))))))
  ;; now create first from 'firsts
  (dotimes (p (1+ size))
    ;; (format t "~&~s  ~s" (aref dp p) (get (aref dp p) 'firsts))
    (setf (aref first p) (get (aref dp p) 'firsts)))
  first)

(defmacro make-state (p j f)
  `(list ,p ,j ,f nil nil))
(defmacro state-p (state) `(car ,state))
(defmacro state-j (state) `(cadr ,state))
(defmacro state-f (state) `(caddr ,state))
(defmacro state-parses (state) `(fourth ,state))
(defmacro state-cws (state) `(fifth ,state))
(defmacro not-statep (x) `(not (integerp (car ,x))))
(defun state-equal (i1 i2)
  (and (= (state-p i1) (state-p i2))
       (= (state-j i1) (state-j i2))
       (= (state-f i1) (state-f i2))))

(defvar *max-n* 20) ;; 40 words in sentence max.
(defvar *state-table* nil)
(defvar *comptable* (make-array (+ *max-n* 2) :initial-element nil))
(defvar *l-vector* (make-array (+ *max-n* 1) :initial-element nil))
(defvar *d-vector* (make-array (+ *max-n* 1) :initial-element nil))
(defvar *w-vector* (make-array (+ *max-n* 1) :initial-element nil))
(defvar *s-vector* (make-array (+ *max-n* 2) :initial-element nil))
(defvar *final-state* (make-state 0 2 0))

;; the state cache is used to keep consing to a minimum.

(defvar *cache-size* 1500)
(defvar *state-cache* nil)
(defvar *cache-pointer* nil)

;; initialize the cache.

(setq *state-cache* nil)
(dotimes (i *cache-size*) (push (make-state nil nil nil) *state-cache*))

(defun get-state (p j f)
  (let ((s (pop *cache-pointer*)))
    (cond (s (setf (state-p s) p
		   (state-j s) j
		   (state-f s) f)
	     s)
	  (t (make-state p j f)))))

;; initialize the *s-vector* on loadup.

(dotimes (i (+ 2 *max-n*))
	 (setf (aref *s-vector* i)
	       (cons nil (make-array (1+ i) :initial-element nil))))

(defun init-parse (ls ds n)
  ;; clear cache cruft.
  (do ((s *state-cache* (cdr s)))
      ((or (null s) (eq s *cache-pointer*)))
      (setf (state-parses (car s)) nil
	    (state-cws (car s)) nil))
  ;; reset cache pointer
  (setq *cache-pointer* *state-cache*)
  ;; clear out all the arrays.
  (dotimes (i (+ *max-n* 2))
	   (setf (aref *comptable* i) nil)
	   (setf (car (aref *s-vector* i)) nil)
	   (dotimes (j (1+ i)) (setf (aref (cdr (aref *s-vector* i)) j) nil)))
  (dotimes (i n)
	   (setf (aref *l-vector* i) (nth i ls))
	   (setf (aref *d-vector* i) (nth i ds))
	   (setf (aref *w-vector* i) (mapcar #'gcar (nth i ds))))
  (setf (aref *l-vector* n) '$)
  (setf (aref *d-vector* n) '($))
  (setf (aref *w-vector* n) '($)))

(defun states (s e &key matching print-parses print-rules)
  (let* ((cpp (sixth *state-table*))
	 (dp (third *state-table*))
	 (gpsg-table (fifth *state-table*))
	 (s-vector (fourth *state-table*))
	 (states (remove-if-not #'(lambda (state)
				    (and (= s (state-f state))
					 (final-p state cpp)))
				(get-state-list s-vector e))))
    (dolist (state states)
	    (unless (and matching (notany #'(lambda (parse) (filt-1 (car parse) matching))
					  (cdr (state-parses state))))
		    (print-state s state dp cpp :failed t)
		    (if print-rules
			(if (null gpsg-table)
			    (error "not a gpsg parse! can't use :print-rules.")
			  (dolist (rule (aref gpsg-table (state-p state)))
				  (format t "~%  rule: ~s" rule))))
		    (if print-parses
			(dolist (parse (cdr (state-parses state)))
				(when (or (null matching) (filt-1 (car parse) matching))
				      (format t "~%")
				      (write parse)))))))
  (values))

(defun earley-parse (g gd words &optional (use-lookahead-if-available t)
		       (print-states nil))
  ;; this algorithm differs from jay earley's in that it only allows k = 0 or k = 1
  ;; (either zero or one token lookahead).  thus a state item is only a triple and not a
  ;; quadruple.  in practice 1 is the most efficient value for k.
  ;;
  ;; to make the algorithm a parser instead of just a recognizer, we add one more piece of
  ;; information to each item, making it a quadruple.  thus an item i is (p, j, f, cw).
  ;; the cw [completed with] is the list of items which allowed the completion operator to
  ;; create this item.
  (cond ((or (not (or (rule-set g 'cfg-table) (rule-set g 'gpsg-table)))
	     (not (rule-set gd 'dictionary)))
	 (format t "~&either ~s is not a cfg-table or a gpsg-table, " g)
	 (format t "or ~s is not a dictionary." gd)
	 nil)
	((member nil (mapcar #'(lambda (w) (dictionary (third (rule-set gd)) w)) words))
	 (format t "~&one or more words in ~s were unknown." words)
	 nil)
	(t
	 (let ((g* (car (third (rule-set g))))
	       (gd* (third (rule-set gd))))
	   (let ((production-table (first g*))
		 (non-terminals (second g*))
		 (cpp (third g*))
		 (dp (fourth g*))
		 (first (if use-lookahead-if-available (fifth g*)))
		 (gpsg-table (sixth g*))
		 (n (length words))
		 (defs (mapcar #'(lambda (w) (dictionary gd* w)) words)))
	     (init-parse words defs n)
	     (setq *last-parse-parses* nil)
	     (add-state *s-vector* 0 0 0 0 *comptable* cpp)
	     ;;
	     ;; main earley loop, i goes from 1 to n.
	     ;;
	     (dotimes (i (1+ n))
		(let ((state-list (get-state-list *s-vector* i)))
		  (if (null (get-state-list *s-vector* i)) (return))
		  ;;
		  ;; process states in order.
		  ;;
		  (do ((states state-list (cdr states)))
		      ((null states))
		      (let ((state (car states)))
			(predictor state i *s-vector* cpp *w-vector* non-terminals
				   production-table first *comptable*)
			(completer state i *s-vector* cpp dp *comptable*)
			(scanner state i *s-vector* cpp *w-vector* *l-vector* *comptable*
				 *d-vector*)))
		  ;;
		  ;; print out all states.
		  ;;
		  (if print-states
		      (dolist (state state-list) (print-state i state dp cpp)))))
	     ;;
	     ;; check for success.
	     ;;
	     ;; if si+1 = |(0,2,0)|, return acceptance.
	     (cond ((and (= 1 (length (get-state-list *s-vector* (1+ n))))
			 (state-equal *final-state*
				      (car (get-state-list *s-vector* (1+ n)))))
		    (setq *state-table* (list (second (rule-set g)) n
					      dp *s-vector* gpsg-table cpp))
		    t)
		   (t (setq *state-table* (list (second (rule-set g)) n dp *s-vector*
						gpsg-table cpp))
		      nil)))))))

(defvar *last-parse-parses* nil)

(defun retrieve-parses (&optional (template '*))
  ;; this function retrieves parses from the last earley parse that match the given template.
  (cond (*last-parse-parses*
	 (filt (cdr *last-parse-parses*) template))
	((eq (first *state-table*) 'cfg-table)
	 (let ((n (second *state-table*))
	       (dp (third *state-table*))
	       (s-vector (fourth *state-table*)))
	   (let ((top-items (get-top-states (first (get-state-list s-vector (1+ n))))))
	     (setq *last-parse-parses*
		   (cons t (mapcan #'(lambda (item)
				       (get-all-parses item dp nil nil))
				   top-items)))
	     (filt (cdr *last-parse-parses*) template))))
	((eq (first *state-table*) 'gpsg-table)
	 (let ((n (second *state-table*))
	       (dp (third *state-table*))
	       (s-vector (fourth *state-table*))
	       (gpsg-table (fifth *state-table*)))
	   (let ((top-items (get-top-states (first (get-state-list s-vector (1+ n))))))
	     (setq *last-parse-parses*
		   (cons t (mapcan #'(lambda (item)
				       (get-all-parses item dp nil gpsg-table))
				   top-items)))
	     (filt (cdr *last-parse-parses*) template))))
	(t (msg t "~%the parse data is not available.~%") nil)))

(defun get-top-states (s)
  (mapcar #'cdr (state-cws (caar (state-cws s)))))

(defun predictor (state i s-vector cpp w-vector non-terminals production-table first comptable)
  (unless (final-p state cpp) ;; if s is a nonfinal state
    (let ((cpj+1 (nth (state-j state) (aref cpp (state-p state)))))
      (when (member cpj+1 non-terminals) ;; and cp(j+1) is a non-terminal
	    ;; then for each q s.t. dq = cp(j+1) add (q,0,i) to si
	    ;; if dq can possible expand into xi.
	(dolist (q (get cpj+1 production-table))
	  (if (or (not first)
		  (member '*empty* (aref first q))
		  (some #'(lambda (w) (member w (aref first q)))
			(aref w-vector i)))
	      (add-state s-vector i q 0 i comptable cpp)))))))

(defun completer (state i s-vector cpp dp comptable)
  (if (final-p state cpp) ;; if s is a final state
      ;; then for each (q,l,g) in sf (after all states have been added to sf) such that
      ;; cq(l+1) = dp, add (q,l+1,g) to si.
      (let* ((d (aref dp (state-p state)))
	     (states (cdr (assoc d (aref comptable (state-f state))))))
	(mapc #'(lambda (sf)
		  (completer-add-state s-vector i (state-p sf) (1+ (state-j sf))
				       (state-f sf) (cons sf state) comptable cpp))
	      states))))

(defun scanner (state i s-vector cpp w-vector l-vector comptable d-vector)
  (when (not (final-p state cpp)) ;; if s is a nonfinal state
	(let ((cpj+1 (nth (state-j state) (aref cpp (state-p state)))))
	  ;; and cp(j+1) is terminal, then if cp(j+1) = xi+1, add (p,j+1,f) to si+1.
	  (if (member cpj+1 (aref w-vector i))
	      (completer-add-state s-vector (1+ i) (state-p state) (1+ (state-j state))
				   (state-f state) (cons state (list cpj+1
								     (aref l-vector i)
								     (aref d-vector i)))
				   comptable cpp)))))

(defun final-p (state cpp)
  ;; a final state is one in which j = (number of symbols on right side of production)
  (= (state-j state) (length (aref cpp (state-p state)))))

(defun add-state-to-comptable (state set-num cj comptable)
  (let ((cell (assoc cj (aref comptable set-num))))
    (when (null cell)
      (setq cell (cons cj nil))
      (push cell (aref comptable set-num)))
    (push state (cdr cell))))

(defun add-state (s-vector set-num p j f comptable cpp)
  (let* ((states (aref s-vector set-num))
	 (sf-vect (cdr states)))
    (unless (find-if #'(lambda (state) (and (= p (state-p state)) (= j (state-j state))
					    (= f (state-f state))))
		     (aref sf-vect f))
      (let* ((state (get-state p j f))
	     (begin-and-end-pointer (car states)))
	(cond ((null begin-and-end-pointer) ;; no state items in state set yet.
	       (push state (aref sf-vect f)) ;; put state in vector for quick lookup.
	       (let ((first-item (cons state nil)))
		 (rplaca states (cons first-item first-item)))
	       (add-state-to-comptable state set-num (nth j (aref cpp p)) comptable))
	      (t (push state (aref sf-vect f)) ;; put state in vector for quick lookup.
		 (let ((new-end (cons state nil))) ;; put state at end of table for
		   ;;general search.
		   (rplacd (cdr begin-and-end-pointer) new-end)
		   (rplacd begin-and-end-pointer new-end))
		 (add-state-to-comptable state set-num (nth j (aref cpp p)) comptable)))))))
  
(defun completer-add-state (s-vector set-num p j f cw comptable cpp)
  ;; when the completion operator creates new states, many of them are equal, since
  ;; an item may be completed in an ambiguous manner.  to record all parses however, we
  ;; need to record each completion.  this function handles such cases.
  (let* ((states (aref s-vector set-num))
	 (sf-vect (cdr states)))
    (let ((prev-state
	   (find-if #'(lambda (state) (and (= p (state-p state)) (= j (state-j state))
					   (= f (state-f state))))
		    (aref sf-vect f))))
      ;; each cw is of the form (a . b) where a is the item this new item was created from,
      ;; and b is the item that caused the creation.
      (if prev-state (push cw (state-cws prev-state))
	(let* ((state (get-state p j f))
	       (begin-and-end-pointer (car states)))
	  (setf (state-cws state) (list cw))
	  (cond ((null begin-and-end-pointer) ;; no state items in state set yet.
		 (push state (aref sf-vect f)) ;; put state in vector for quick lookup.
		 (let ((first-item (cons state nil)))
		   (rplaca states (cons first-item first-item)))
		 (add-state-to-comptable state set-num (nth j (aref cpp p)) comptable))
		(t (push state (aref sf-vect f)) ;; put state in vector for quick lookup.
		   (let ((new-end (cons state nil)))
		     ;; put state at end of table for general search.
		     (rplacd (cdr begin-and-end-pointer) new-end)  
		     (rplacd begin-and-end-pointer new-end))
		   (add-state-to-comptable state set-num (nth j (aref cpp p)) comptable))))))))

(defun get-state-list (s-vector set-num)
  (car (car (aref s-vector set-num))))

(defun get-all-parses (item dp trail gpsg-table)
  (cond ((state-parses item) (cdr (state-parses item))) ;; memoized.
	((member item trail)
	 (if gpsg-table nil ;; doing gpsg parse.  no recursion allowed.
	   (list (format nil "recursive-~d"
			 (position item trail :test #'eq)))))
	((null (state-cws item)) ;; empty category
	 (if gpsg-table
	     (gpsg-handle '(nil) gpsg-table (state-p item))
	   (list (list (aref dp (state-p item))))))
	(t (do ((cws-and-parses (mapcar #'(lambda (cw) (cons cw (list nil)))
					 (state-cws item))
				extended-cws-and-parses)
		(extended-cws-and-parses nil nil))
	       ((null (car (first cws-and-parses)))
		(let ((results nil))
		  (dolist (cw-and-parses cws-and-parses)
		    (if gpsg-table
			(setq results (nconc results
					     (gpsg-handle (cdr cw-and-parses)
							  gpsg-table (state-p item))))
		      (dolist (parse (cdr cw-and-parses))
			      (push (cons (aref dp (state-p item)) parse)
				    results))))
		  (setf (state-parses item) (cons t results))
		  results))
	       (dolist (cw-and-parses cws-and-parses)
		 (let ((extended-parses nil))
		   (cond ((not-statep (cdr (car cw-and-parses)))
			  ;; lexical item is sub-tree.
			  (dolist (sub-tree (lexicalize-item (cdr (car cw-and-parses))))
			    (dolist (unfinished-parse (cdr cw-and-parses))
			      (push (cons sub-tree unfinished-parse) extended-parses)))
			  (dolist (cw (or (state-cws (car (car cw-and-parses))) '(nil)))
				  (push (cons cw extended-parses)
					extended-cws-and-parses)))
			 (t
			  (dolist (sub-tree (get-all-parses (cdr (car cw-and-parses)) dp
							    (cons item trail)
							    gpsg-table))
			    (dolist (unfinished-parse (cdr cw-and-parses))
			      (push (cons sub-tree unfinished-parse) extended-parses)))
			  (dolist (cw (or (state-cws (car (car cw-and-parses))) '(nil)))
				  (push (cons cw extended-parses)
					extended-cws-and-parses))))))))))


(defun lexicalize-item (item)
  ;; item = (category word (word-dictionary-definitions))
  (let ((result nil))
    (dolist (definition (third item))
	    (when (eq (gcar definition) (car item)) ;; correct lexical category
		  (push (list (car item) (second item)) ;; charles hack for semantics lab. see carl's original
			result)))
    result))

(defun print-state (i state dp cpp &key failed)
  (if failed
      (format t (if (cdr (state-parses state))
		    "~%_ ~d  ~d   ~s ==>~{ ~s~} .~{ ~s~}"
		    "~%* ~d  ~d   ~s ==>~{ ~s~} .~{ ~s~}")
	      i (state-f state) (aref dp (state-p state))
	      (subseq (aref cpp (state-p state)) 0 (state-j state))
	      (subseq (aref cpp (state-p state)) (state-j state)))
      (format t "~%_ ~d  ~d   ~s ==>~{ ~s~} .~{ ~s~}"
	      i (state-f state) (aref dp (state-p state))
	      (subseq (aref cpp (state-p state)) 0 (state-j state))
	      (subseq (aref cpp (state-p state)) (state-j state)))))

(defun remove-tag (sentence) (if (eq (car sentence) '*) (cdr sentence) sentence))

(defvar *gram* nil)
(defvar *dict* nil)
(defun p (sentence &key (template '*) grammar dictionary (use-lookahead t)
	  print-states file sentence-set)
  (if grammar (setq *gram* grammar))
  (if dictionary (setq *dict* dictionary))
  (cond ((null (rule-set *dict* 'dictionary))
	 (format t "~&please specify a dictionary.")
	 (format t "~&(p sentence &key (template '*) grammar dictionary ")
	 (format t "(use-lookahead t) print-states) file sentence-set"))
	((and sentence-set (null (rule-set sentence-set 'sentences)))
	 (format t "~&~s is not a sentences rule set." sentence-set))
	((null (or (rule-set *gram* 'gpsg-table)
		   (rule-set *gram* 'cfg-table)))
	 (format t "~&please specify either a cfg-table or a gpsg-table.")
	 (format t "~&(p sentence &key (template '*) grammar dictionary ")
	 (format t "(use-lookahead t) print-states) file sentence-set"))
	(t (let ((sents (if sentence-set
			    (mapcar #'listify (third (rule-set sentence-set)))
			    (list (listify sentence)))))
	     (cond
	       (file
		(with-open-file (out file :direction :output :if-exists :new-version)
		  (format out "~%~%~%*** parse results ***~%")
		  (mapc #'(lambda (sent)
			    (format out "~%~%sentence: ~{~s ~}" sent)
			    (let ((ok (earley-parse *gram* *dict* (remove-tag sent)
						    use-lookahead print-states)))
			      (msg t "~&done cfg parse.")
			      (let ((results (and ok (retrieve-parses template))))
				(msg t "~&done parse retrieval")
				(format out "~%number of parses: ~s~%" (length results))
				(dolist (s results)
					(format out "~%")
					(write s :stream out)
					(format out "~%")))))
			sents))
		nil)
	       (sentence-set
		(mapcar #'(lambda (sent)
			    (let ((ok (earley-parse *gram* *dict* (remove-tag sent)
						    use-lookahead print-states)))
			      (msg t "~&done cfg parse.")
			      (let ((result (cons sent (and ok (retrieve-parses template)))))
				(msg t "~&done parse retrieval")
				result)))
			sents))
	       (t
		(let ((ok (earley-parse *gram* *dict* (remove-tag (car sents))
					use-lookahead print-states)))
		  (msg t "~&done cfg parse.")
		  (let ((result (and ok (retrieve-parses template))))
		    (msg t "~&done parse retrieval")
		    result))))))))

;;;
;;; random junk
;;;

(defun listify (sent)
  (if (listp sent) sent
      (read-from-string
	(concatenate 'string "(" sent ")"))))

(defvar *verbose-p* t)

(defun msg (&rest data)
  (if *verbose-p* (apply #'format data)))

(defun verbose (&key (set 0))
  (if (eql set 0)
      (setq *verbose-p* (not *verbose-p*))
      (setq *verbose-p* set))
  *verbose-p*)

(defun s< (a b)
  (let ((sa (cond ((stringp a) (subseq a 0 (position #\space a)))
		  ((symbolp a) (string a))
		  ((symbolp (car a)) (string (car a)))
		  (t (string (caar a)))))
	(sb (cond ((stringp b) (subseq b 0 (position #\space b)))
		  ((symbolp b) (string b))
		  ((symbolp (car b)) (string (car b)))
		  (t (string (caar b))))))
    (string< sa sb)))

;;;
;;; gpsg specific stuff.
;;;

(defmacro feature-p (spec feature) `(member ,feature ,spec))
(defmacro rhs (rule) `(cddr ,rule))
(defmacro lhs (rule) `(car ,rule))
(defmacro feature-value (spec feature)
  `(second (member ,feature ,spec)))

(defun expand-gpsg (rule-list)
  (let ((slash-categories nil)
	(non-terminals nil)
	(rules-for-expansion nil)
	(rules nil))
    (flet ((learn-slashes (spec)
	     (if (feature-p spec '/)
		 (let ((slash-value (feature-value spec '/)))
		   (if (and slash-value (not (variable-p slash-value)))
		       (setq slash-categories
			     (adjoin (gcar slash-value) slash-categories))))))
	   (expand-rule (rule)
             (let ((new-lhs (append (lhs rule) '(/ ?slash)))
		   (old-rhs (rhs rule)))
	       (do ((rhs-rest old-rhs (rest rhs-rest))
		    (n 0 (1+ n)))
		   ((null rhs-rest))
		   (let ((spec (car rhs-rest)))
		     (unless (or (feature-p spec '/) (not (member (car spec) non-terminals)))
		       (let ((new-rhs (copy-list old-rhs)))
			 (setf (nth n new-rhs) (append spec '(/ ?slash)))
			 (push (list* new-lhs '==> new-rhs) rules))))))))
       (setq rules (mapcar #'(lambda (rule)
			       ;; learn the non-terminals.
			       (setq non-terminals (adjoin (gcar (lhs rule)) non-terminals))
			       (let ((lhs (glist (lhs rule)))
				     (rhs (mapcar #'glist (rhs rule))))
				 (list* lhs '==> rhs)))
			   rule-list))
       (dolist (rule rules)
	       ;; don't expand rule with slashes if it has a slash specification
	       ;; in the lhs.
	       (unless (feature-p (lhs rule) '/)
		       (push rule rules-for-expansion))
	       ;; learn the possible slash categories.
	       (learn-slashes (lhs rule))
	       (dolist (spec (rhs rule)) (learn-slashes spec)))
       ;; now expand rule set to allow one extra slashed category per rule.
       (dolist (rule rules-for-expansion) (expand-rule rule))
       (format t "~%original rule set length: ~d~%with new slash rules added: ~d"
	       (length rule-list) (length rules))
       (let ((final-rules (slashify-rules rules slash-categories)))
	 (format t "~%after slash instantiation: ~d" (length final-rules))
	 final-rules))))

(defun slashify-rules (rules slash-categories)
  ;; translate rules like ((s / np) => np (vp / np))
  ;;                 into s/np ==> np vp/np.
  (let ((new-rules nil))
    (dolist (rule rules)
      ;; first find all the instances of "/ ?x".
      (let ((slash-variables nil)
	    (expanded-rules (slashify-rule rule 'qqq '(qqq) t)))
	(labels ((get-slash (spec)
		    (if (feature-p spec '/)
			(let ((val (gcar (feature-value spec '/))))
			  (if (variable-p val)
			      (setq slash-variables
				    (adjoin val slash-variables)))))))
		(get-slash (lhs rule))
		(dolist (spec (rhs rule)) (get-slash spec)))
	(dolist (variable slash-variables)
		(setq expanded-rules
		      (mapcan #'(lambda (rule)
				  (slashify-rule rule variable slash-categories))
			      expanded-rules)))
	(setq new-rules (nconc new-rules expanded-rules))))
    new-rules))

(defun slashify-rule (rule variable categories &optional constants)
  (labels ((expand-spec (spec variable category)
	     (let ((value (feature-value spec '/)))
	       (if (eq variable value)
		   (cons (make-slashed-symbol (car spec) category)
			 (cdr spec))
		 (if (and value constants (not (variable-p value)))
		     (cons (make-slashed-symbol (car spec) (gcar value))
			   (cdr spec))
		   spec)))))
	  (mapcar #'(lambda (category)
		      (list* (expand-spec (lhs rule) variable category)
			     '==>
			     (mapcar #'(lambda (spec)
					 (expand-spec spec variable category))
				     (rhs rule))))
		  categories)))

(defun make-slashed-symbol (symbol category)
  (let ((new-symbol (read-from-string (format nil "~s/~s" symbol category))))
    (setf (get new-symbol 'unslashed) symbol)
    new-symbol))

(defun remove-slash (symbol)
  (or (get symbol 'unslashed) symbol))

(defun slashed-p (symbol) (get symbol 'unslashed))

(defun get-gpsg (lst)
  (let ((rule-lst nil)
	(gpsg-lst nil))
    (dolist (r lst)
      (let ((rule (mapcar #'gcar r)))
	(setq rule-lst (adjoin rule rule-lst :test #'equal))
	(let ((cell (assoc rule gpsg-lst :test #'equal)))
	  (unless cell
	    (setq cell (cons rule nil))
	    (push cell gpsg-lst))
	  (mapl #'(lambda (l)
		    (if (symbolp (car l))
			(rplaca l (remove-slash (car l)))
		      (rplaca (car l) (remove-slash (caar l)))))
		r)
	  (push r (cdr cell)))))
    ;; we now have to filter out some obviously impossible rules.
    (setq rule-lst (gpsg-filter-rules rule-lst))
    (format t "~%rules filtered down to: ~d" (length rule-lst))
    (setq gpsg-lst (delete-if #'(lambda (g) (not (member (car g) rule-lst))) gpsg-lst))
    (list rule-lst gpsg-lst)))

(defun gpsg-filter-rules (rule-set)
  (do ((removed-some t removed-some)
       (rule-set rule-set rule-set))
      ((not removed-some) (delete nil rule-set))
    (setq removed-some nil)
    (let ((category-list nil))
      (dolist (r rule-set)
	(if r (setq category-list (adjoin (lhs r) category-list))))
      (do ((rs rule-set (cdr rs)))
	  ((null rs))
	  (when (and (car rs)
		     (notevery #'(lambda (c) (or (not (slashed-p c))
						 (member c category-list)))
			       (rhs (car rs))))
		(setq removed-some t)
		(rplaca rs nil))))))

(defun glist (s) (if (symbolp s) (list s) s))
(defun gcar (s) (if (symbolp s) s (car s)))
(defun variable-p (symbol)
  (and (symbolp symbol) (char= #\? (aref (string symbol) 0))))

(defun gpsg-handle (parses gpsg-table rule)
  (let ((resulting-parses nil))
    (dolist (parse parses)
      (let ((rules (aref gpsg-table rule)))
	(mapc #'(lambda (rule)
		  (let ((result (gpsg-handler parse rule)))
		    (if result (push result resulting-parses))))
	      rules)))
    resulting-parses))

(defun gpsg-handler (parse rule)
  ;; use the given gpsg rule and parse tree to expand the head of the parse tree.
  (let ((unification (list '**unify)))
    (if (some #'(lambda (parse-node rule-spec)
		  (gpsg-compare (car parse-node) (glist rule-spec) unification))
	      parse (cddr rule))
        (progn
	  (msg t "~%rule failure: ~{ ~s~}~%sub trees: ~{ ~s~}" rule
	       (mapcar #'car parse))
	  nil)
      (cons (gpsg-expand (glist (car rule)) unification) parse))))
      
(defun gpsg-compare (parse-category rule-spec unification-table)
  ;; return t if the given parse-category can not match the rule-spec under the
  ;; unification table unification-table.
  (do ((rest-spec (cdr rule-spec) (cddr rest-spec)))
      ((null rest-spec) nil)
      (let ((f-value (cadr rest-spec))
	    (f-name (car rest-spec)))
	(let ((parse-value (feature-value parse-category f-name)))
	  (cond ((and (null parse-value) f-value)
		 ;; tree node leaves feature unspecified.
		 (if (obligatory-p f-name) (return t)))
		;; handle "/ nil"
		((and (null parse-value) (null f-value) (eq f-name '/)))
		((variable-p f-value) ;; variable
		 (let ((unification-value (cdr (assoc f-value (cdr unification-table)))))
		   (if unification-value
		       (unless (gpsg-eq parse-value unification-value
					unification-table)
			       (return t))
		     ;; value for variable not determined yet.
		     (rplacd unification-table
			     (cons (cons f-value parse-value)
				   (cdr unification-table))))))
		((gpsg-eq f-value parse-value unification-table))
		(t (return t)))))))

(defun obligatory-p (symbol)
  (and (symbolp symbol) (char= #\+ (aref (string symbol) 0))))

(defun gpsg-expand (category unification-table)
  (cons (car category)
	(if (some #'variable-p (cdr category))
	    (do ((result nil result)
		 (spec (cdr category) (cddr spec)))
		((null spec) (nreverse result))
		(if (variable-p (cadr spec))
		    (let ((value (cdr (assoc (cadr spec) (cdr unification-table)))))
		      (if value (setq result (list* value (car spec) result))))
		  (setq result (list* (cadr spec) (car spec) result))))
	  (cdr category))))

(defun gpsg-eq (a b unification-table)
  ;; are these values equal, where a is a parse-category value and b is
  ;; a rule-spec value.
  (cond ((equalp a b))
	((or (eq a '*) (eq b '*)))
	((and (symbolp a) (symbolp b)) nil)
	((and (symbolp a) (listp b)) (eql a (car b)))
	((and (symbolp b) (listp a)) (eql b (car a)))
	(t (not (gpsg-compare a b unification-table)))))


