;;; -*- mode: lisp; package: gpsg; base: 10; syntax: common-lisp -*-
;;; example: john saw mary.
;;; syntax: (s (np (np-pro (name john))) 
;;;            (vp (advp) (v+args (v2+tns saw) (np (np-pro (name mary)))))))
;;; semantics: (see :agent john :patient mary :tense past)

;this is the wrong order according to guy steele, but if in-package isn't the
;first form in the file, lucid gives a load error.
; (in-package 'gpsg :use '(lisp))

(provide 'interpreter)

(defvar *lexical-semantics* '(()) "word meaning database")
(defvar *phrasal-semantics* '(()) "phrase meaning database")

;; following var is set to whatever you  want, as is true of the grammar
(defvar *syntax-lexicon* 'g1 "the syntax lexicon rule set")
;; following var is set to whatever you named your grammar via
;; (add-rule-list ..) while the next var is what you named your
;; grammar via (create-cfg-table ...)
(defvar *syntax-grammar* 'semg "the syntax grammar rule set")
(defvar *syntax-cfg* 'semg1 "the syntax cfg")

(defvar *rule-semantics-file* "new-rules" "the rule semantics defns")

(defvar *word-semantics-file* "new-words" "the word semantics defns")

(defvar *show-syntax?* 1 "show syntax of assertions and queries?")
(defvar *show-semantics?* 1 "show semantics of assertions and queries?")

;;; this function takes a syntactical parse-tree and returns 
;;; a thematic role frame.
;;; here is how it works.  it recursively maps itself (via mapcar) 
;;;    1. if the phrase is a single word, it will be in the form
;;;       (category word), e.g.,
;;;         (ate v2+tns)
;;;       so the car of this phrase is the word's syntactic category and 
;;;       the cadr is the word itself.  this is used to access 
;;;       the doubly-indexed
;;;       *lexical-semantics* via the access function word-semantics
;;;       this representation is accessed via lookup2 
;;;
;;;       this returns a lambda-form for the word (a thematic slot-and-filler
;;;       representation
;;;
;;;    2. if the element is a phrase, then the access function on 
;;;       phrase-semantics returns a lambda-form whose application
;;;       is to be associated with some ``set'' of objects (its denotation)
;;;       in some cases, this is a higher-order function, e.g., a
;;;       verb phrase's meaning is generally itself a lambda form; hence
;;;       the use of a back-quoted macro form as the object returned.
;;;
;;;    3. the function rule-apply has the job of actually applying
;;;       the resulting lambda objects to their arguments.
;;;       the rather ugly form of the application, which should really
;;;       just be something like (apply head args), is the result
;;;       of corresponding ugliness in common lisp.  so it goes.
;;;
;;;    4. eventually, what is returned is a database template that is
;;;       matched against a database of facts stored in the global 
;;;       variable *db*
;;;       (see code in the file
;;;       match-allowing-extras).  this will either add a new fact
;;;       to the db, or match a query against it.


(defun phrase-semantics (phrase)
  (cond  ((atom (cadr phrase)) 
	 (word-semantics (cadr phrase) (car phrase)))
	(t (rule-apply (rule-semantics (car phrase)
                                       (mapcar #'car (cdr phrase)))
                       (mapcar #'phrase-semantics (cdr phrase))))))

(defun rule-apply (head args)
  (let ((result (apply head args)))
    (if (and (consp result)
             (eq (car result) 'lambda))
      (eval (list 'function result))
      result)))
;;; top-level read-eval-print-loop.  note that one can re-evaluate
;;; a semantic rule at any time just by calling add-rule-sem again,
;;; but that to use a new grammar one must build the grammar again
;;; by calling (init-syntax)


(defun top-level ()
  (declare (special *show-syntax?*))
  (declare (special *show-semantics?*))
  (declare (ignore semantics))
  (when (y-or-n-p "~&shall i clear the database? (y or n) ")
    (db-clear))
  (loop
    (format t "~&sem-interpret>")
    (let ((sent (read-line)))
      (cond ((equal sent "quit") (return t))   ;; type "quit" to get out
	    ((eq (length sent) 0) (return t))  ;; or just a carriage return
	    ((eq (aref sent 0) #\( )
	     (format t "~&= ~a" (eval (read-from-string sent))))
	    (t (let ((parse (car (p sent))))
		 (if *show-syntax?* (format t "~1% syntax: ~a" parse))
		 (phrase-semantics parse)))))))


;;; semantic database.  looks up the interpretive procedure associated
;;; with a word by using a doubly-indexed association list, indexed
;;; by the head word and then its grammatical category
;;; example: "eat" is indexed under "eat" and v1 (verb in category 1)
;;; then the function actually runs the associated lambda procedure
;;; associated, if any. 
;;; example: (lookup2 'who 'pronp+wh *lexical-semantics*) returns
;;; ?who
;;;

(defun word-semantics (word sense)
  (let ((x (lookup2 word sense *lexical-semantics*)))
    (if (and (consp x)
             (eq (car x) 'lambda))
      (eval (list 'function x))
      x)))


(defun rule-semantics (head args)
  (let ((x (lookup2 head args *phrasal-semantics*)))
    (if (and (consp x)
             (eq (car x) 'lambda))
      (eval (list 'function x))
      x)))

;;; this function just builds a lexical semantics database, using a doubly
;;; indexed association list: indexed
;;; by the word and then its syntactic category.  add2 is the insertion
;;;  routine.
;;; then it calls a function to build the right dictionary entry
;;; for the earley parser
;;;
(defun add-word-semantics (word sense semantics)
  (add2 word sense semantics *lexical-semantics*)
  (add-word-to-syntax word sense))



;;;  grammar rule format
;;;  rule is defined as a pair of (cfg-rule, semantics defined by lambda proc)
;;; example: (add-rule-sem '(s ==> np vp) '(lambda (np vp) (funcall vp np)))

(defun lhs-of (rule)
  (car rule))

(defun rhs-of (rule)
  (cddr rule))

(defun add-rule-sem (rule semantics)
  (add2 (lhs-of rule) (rhs-of rule) semantics *phrasal-semantics*)
  (add-rule *syntax-grammar* rule))

;;; database utilities
;;; a doubly-indexed association list.  access routine is lookup2
;;; while the insertion is done by add2

(defun lookup2 (key1 key2 db)
  (cdr (assoc key2 (cdr (assoc key1 db :test #'equal)) :test #'equal)))

(defun add2 (key1 key2 data db)
  (let ((bucket1 (assoc key1 db :test #'equal)))
    (cond ((null bucket1)
	   (push `(,key1 . ((,key2 . ,data))) (cdr db)))
	  (t (let ((bucket2 (assoc key2 (cdr bucket1) :test #'equal)))
	       (cond ((null bucket2)
		      (push `(,key2 . ,data) (cdr bucket1)))
		     (t (cond ((equal (cdr bucket2) data))
			      (t (format t "~&redefining semantics (~s ==> ~s) to be ~s"
					 key1 key2 data)
				 (setf (cdr bucket2) data))))))))))


;;; syntax interface designed to build proper
;;; dictionary for earley parser.  uses earley parser add-rule
;;; function using globally defined variables as shown


(defun add-word-to-syntax (word sense)
  (declare (special *syntax-lexicon*))
  (add-rule *syntax-lexicon* (list word sense)))


;;; discourse database

(defvar *db* '(()))

(defun db-clear ()
  (declare (special *db*))
  (setf (cdr *db*) nil))

(defun db-add (fact)
  (declare (special *db*))
  (setf (cdr *db*) (cons fact (cdr *db*)))
  fact)

;; returns one value
(defun db-match-1 (query) 
  (declare (special *db*))
  (dolist (datum (cdr *db*))
    (let ((e (match-allowing-extras query datum)))
      (when e (return e))))
  nil)

;; returns n values
(defun db-match-n (query)
  (declare (special *db*))
  (remove-if-not #'identity
		 (mapcar #'(lambda (datum)
			     (match-allowing-extras query datum))
			 *db*)))

;;; declaratives 

(defun process-sentence (sentence)
  (declare (special *show-semantics?*))
  (if *show-semantics?* (format t "~2% semantics: ~s~%" sentence))
  (db-add sentence)
  (format t "~%ok."))

;;; questions

(defun wh-question (wh question)
  (declare (special *show-semantics?*))
  (if *show-semantics?* (format t "~& semantics: wh-question ~a ~s~%" wh question))
  (let ((envs (db-match-n question)))
    (cond ((null envs) (format t "~&i don't know."))
	  (t (let ((answer (remove-duplicates
			     (mapcar #'(lambda (e) (instantiate wh e))
				     envs))))
	       (say-response answer)
	       (if *show-semantics?*
		   (format t "~& answer semantics: ~s~%" answer)))))))

(defun y/n-question (question)
  (declare (special *show-semantics?*))
  (if *show-semantics?* (format t "~& semantics: yes/no-question ~s~%" question))
  (let ((e (db-match-n question)))
    (cond ((null e) (format t "~&i don't think so."))
	  (t (format t "~&yes.")))))

;;;

(defun add-feature (frame feature value)
  `(,@frame ,feature ,value))


;;; a list of tests
(defvar *semantics-tests* '("john saw mary"))
                            

(defun trace-semantics ()
  (trace phrase-semantics)
  (trace rule-semantics)
  (trace word-semantics))

(defun untrace-semantics ()
  (untrace phrase-semantics)
  (untrace rule-semantics)
  (untrace word-semantics))

(defun test-semantics ()
  (dolist (sent *semantics-tests*)
    (format t "~&sentence: ~s" sent)
    (format t "~&syntax: ~s" (p sent))
    (format t "~&semantics: ~s~% " (s sent))))

;;; testing function: calls parser and semantics on sentence
(defun s (sentence)
  (let ((parses (p sentence)))
    (cond ((= (length parses) 1)
	   (phrase-semantics (car parses)))
	  (t (mapcar #'phrase-semantics parses)))))

;;; printing: prints out lexical and phrase semantics rules in a more
;;; readable form
(defun list-semantics ()
  (declare (special *lexical-semantics*))
  (declare (special *phrasal-semantics*))
  (let ((semantics nil))
    (dolist (head-frame (cdr *phrasal-semantics*))
      (dolist (body (cdr head-frame))
	(push `(,`(,(car head-frame) ==> ,@(car body)) ,(cdr body)) semantics)))
    (setq semantics (sort semantics #'(lambda (l1 l2)
					(cond ((string< (symbol-name (caar l1))
							(symbol-name (caar l2)))
					       t)
					      ((string> (symbol-name (caar l1))
							(symbol-name (caar l2)))
					       nil)
					      ((string= (symbol-name (caar l1))
							(symbol-name (caar l2)))
					       (string< (symbol-name (caddar l1))
							(symbol-name (caddar l2))))))))
    (dolist (rule semantics)
      (format t "~&~a : ~a ~& " (first rule)
	      (cond ((eq (second rule) #'identity)
		     'identity)
		    (t (second rule))))))
  (let ((words nil))
    (dolist (head-frame (cdr *lexical-semantics*))
      (dolist (body (cdr head-frame))
	(push `(,(car head-frame) ,@(car body)) words)))
    (setq words (sort words #'(lambda (l1 l2)
				(cond ((string< (symbol-name (car l1))
						(symbol-name (car l2)))
				       t)
				      ((string> (symbol-name (car l1))
						(symbol-name (car l2)))
				       nil)
				      ((string= (symbol-name (car l1))
						(symbol-name (car l2)))
				       (string< (symbol-name (cdr l1))
						(symbol-name (cdr l2))))))))
    (dolist (word words)
      (format t "~&~a" word))))
      

;;; answering wh questions

(defun say-response (response)
  (cond ((stringp response)
	 (format t "~&~a" response))
	((null (second response))
	 (say-pp-or-np (first response)))
	(t (say-conjunct response))))

;;; generate a conjunct as output

(defun say-conjunct (responses)
  (say-pp-or-np (car responses))
  (dolist (response (cdr responses))
    (format t " and ")
    (say-pp-or-np response)))

(defun say-pp-or-np (pp-or-np)
  (cond ((and (listp pp-or-np) (eq (car pp-or-np) :relation))
	 (format t "~a " (second pp-or-np))
	 (say-np (getf pp-or-np :location)))
	(t (say-np pp-or-np))))

;;; simple generation routine for nps, checking for a/the and mass
;;; noun properties

(defun say-np (np)
  (cond ((atom np) (format t "~a" np))
	(t (if (getf (cdr np) :definite?)
	       (format t "the ") 
	       (unless (equal (getf (cdr np) :number) 'mass)
		 (format t "a ")))
	   (say-adj (getf (cdr np) :mod))
	   (format t "~a" (car np)))))

;;; spit out a list of adjectives
;;;

(defun say-adj (adj)
  (cond ((null adj))
	((eq (first adj) :color)
	 (format t "~a " (second adj))
	 (say-adj (getf adj :mod)))
	(t (format t "~a " (first adj))
	   (say-adj (getf (cdr adj) :mod)))))

;;; initialize syntax and parse a simple sentence
;;; so dictionary and grammar are initialized.


(defun init-syntax ()
  (declare (special *syntax-lexicon*))
  (declare (special *syntax-grammar*))
  (declare (special *syntax-cfg*))
  (declare (special *rule-semantics-file*))
  (declare (special *word-semantics-file*))
  (remove-rule-set *syntax-lexicon*)
  (remove-rule-set *syntax-grammar*)
  (add-rule-set *syntax-lexicon* 'dictionary)
  (add-rule-set *syntax-grammar* 'cfg)
  (load *rule-semantics-file*)
  (load *word-semantics-file*)
  (remove-rule-set *syntax-cfg*)
  (create-cfg-table *syntax-cfg* *syntax-grammar* 'root 1)
  (p "john saw mary" :grammar *syntax-cfg* :dictionary *syntax-lexicon*))

;;; -*- mode: lisp; package: user; syntax: common-lisp; base: 10 -*-
;;; match-allowing-extras.lisp
;;; public functions: match-allowing-extras, instantiate, combine,
;;;   make-binding-list, 
;;;   uniqueify
;;; nb nil is a binding list that represents failure of a match

;this is the wrong order, but lucid gives a load warning if in-package isn't
;the first line of a file.
; (in-package 'gpsg)

(provide 'match-allowing-extras)

(defun match-allowing-extras (p d &optional (b (make-binding-list)))
  (let ((b (internal-match p d b)))
    (values b (not (equal b (failed-binding-list))))))

(defun internal-match (p d b)
  (cond ((binding-list-failure? b) b)
        ((or (atom p) (atom d))
         (match-pred p d b))
        (t (loop (setf b (match-pred (car p) (car d) b))
		 (unless b (return nil))
		 (setf p (cdr p))
		 (setf d (cdr d))
		 (cond ((variable? p) ;; the (xxx . ?v) case
			(return (match-pred p d b)))
		       ((or (null p) (null d))
;;;			(if (and (null p) (null d)) ;change for match-allowing-extras
			(if (null p)
			    (return b)
			  (return (failed-binding-list)))))))))

(defun match-pred (p d b)
  (cond ((equal p d)
         b)
        ((and (variable? p) (bound? p b))
         (internal-match (lookup-variable p b) d b))
        ((and (variable? d) (bound? d b))
         (internal-match p (lookup-variable d b) b))
        ((and (variable? p) (freefor? d p b))
         (let ((new-b (copy-binding-list b)))
           (bind-variable p d new-b)
	   new-b))
        ((and (variable? d) (freefor? p d b))
         (let ((new-b (copy-binding-list b)))
           (bind-variable d p new-b)
	   new-b))
        (t
         (if (or (atom p) (atom d))
             (if (equal p d)
                 b
                 (failed-binding-list))
             (internal-match p d b)))))

(defun instantiate (p b)
  (cond ((null p) nil)
        ((atom p) 
         (cond ((variable? p)                   
                (let ((p-value (lookup-variable p b)))
                  (cond ((null p-value) p)
                        ((variable? p-value)
                         (instantiate p-value b))
                        (t p-value))))
               (t p)))          
        (t (cons (instantiate (car p) b)
                 (instantiate (cdr p) b)))))

(defun freefor? (exp var b)
  (cond ((variable? exp)
         (cond ((equal exp var) nil)
               (t (freefor? (lookup-variable exp b) var b))))
        ((atom exp) t)
        (t (and (freefor? (car exp) var b)
                (freefor? (cdr exp) var b)))))

;;; binding lists

(defun print-binding-list (binding-list stream print-level)
  (declare (ignore print-level))
  (format stream "{")
  (mapc #'(lambda (pair) (format stream "~s~s " (car pair) (cdr pair)))
	(binding-list-bindings binding-list))
  (format stream "}"))

(defstruct (binding-list (:print-function print-binding-list) (:copier nil)) bindings)

(defun bound? (variable binding-list)
  (assoc variable (binding-list-bindings binding-list)))

(defun lookup-variable (variable binding-list)
  (let ((bucket (assoc variable (binding-list-bindings binding-list))))
    (if bucket (cdr bucket))))

(defun bind-variable (variable value binding-list)
  (push (cons variable value) (binding-list-bindings binding-list)))

(defun binding-list-domain (binding-list)
  (mapcar #'car (binding-list-bindings binding-list)))

(defun binding-list-failure? (binding-list)
  (null binding-list))

(defun copy-binding-list (binding-list)
  (make-binding-list :bindings
		     (copy-alist (binding-list-bindings binding-list))))

(defun failed-binding-list ()
  nil)

(defun valid-env? (env) env) ;; 4-july-88

(defun combine (bl1 bl2) ;; unsure about this fcn
  (let ((bl3 (copy-binding-list bl1))
        (vars-in-bl1 (binding-list-domain bl1)))
    (block match-up
      (dolist (var-value (binding-list-bindings bl2)) ;; cvted to cl 4-jul-88
	(let ((var (car var-value))
	      (value (cdr var-value)))
	  (cond ((member var vars-in-bl1 :test #'equal)
		 (unless (equal (lookup-variable var bl1)
				(lookup-variable var bl2))
			 (return-from match-up (failed-binding-list))))
		(t (bind-variable var value bl3)))))
      bl3)))

;;; variables

(defun variable? (x)
  (and (symbolp x)
       (char= (char (symbol-name x) 0) #\?)))

(defun add-suffix-number-to-variable (variable suffix-number)
  (intern (coerce (append (coerce (symbol-name variable) 'list)
			  (coerce (format nil "-~s" suffix-number) 'list))
		  'string)))

;;; hack to get unique variable names

(defvar *uniqueify-instance-count* 0)

(defun reset-instance-count ()
  (declare (special *uniqueify-instance-count*))
  (setq *uniqueify-instance-count* 0))

(defun uniqueify (exp)
  (declare (special *uniqueify-instance-count*))
  (incf *uniqueify-instance-count*)
  (add-suffix-number-to-variables-in-tree exp *uniqueify-instance-count*))

(defun add-suffix-number-to-variables-in-tree (exp suffix)
  (cond ((variable? exp)
         (add-suffix-number-to-variable exp suffix))
        ((consp exp)
         (cons (add-suffix-number-to-variables-in-tree (car exp) suffix) 
               (add-suffix-number-to-variables-in-tree (cdr exp) suffix)))
        (t exp)))

;;; rules

(defun print-rule (rule stream print-level)
  (declare (ignore print-level))
  (format stream "[~s=>~s/~s"
	  (rule-lhs rule)
	  (rule-rhs rule)
	  (rule-bindings rule))
  (cond ((eq (rule-type rule) 'normal)
	 (format stream "]"))
	(t (format stream " ~s]" (rule-type rule)))))

;;;(defstruct (rule (:print-function print-rule))
(defstruct rule ;; without the print-function 11-may-88
  lhs 
  rhs
  (bindings (make-binding-list))
  (type 'normal)
  (priority 1))

(defvar *sample-rule-set-1* (list (make-rule :lhs '((if ?a ?b) ?a) :rhs '?b)
				(make-rule :lhs '((or ?a ?b) (not ?a)) :rhs '?b)
				(make-rule :lhs '((or ?a ?b) (not ?b)) :rhs '?a)
				(make-rule :lhs '((not (not ?a))) :rhs '?a)
				(make-rule :lhs '((and ?a ?b)) :rhs '?a)
				(make-rule :lhs '((and ?a ?b)) :rhs '?b)))

;;; added 4-july-88

(defun make-env () (make-binding-list))

(defun same-env? (e1 e2)
  (let ((dom1 (binding-list-domain e1))
	(dom2 (binding-list-domain e2)))
    (and (= (length dom1) (length dom2))
	 (every #'(lambda (variable)
		    (equal (lookup-variable variable e1)
			   (lookup-variable variable e2)))
		dom1))))

(defun same-pattern? (p1 p2)
  (equal p1 p2))
		    
(defstruct conflict-set-element rule env datum)

(defun filter (pred seq) (remove-if-not pred seq))

(defun combine-env (env1 env2) (combine env1 env2))


