
//=============================================================================

// This module handles the interface between C and LISP
//
// Expectations and Limits:
//
// 1) It will run the LISP parser by executing "mlisp -s air_control.lisp"
//    So "mlisp" must be in the executable search path.
//
// 2) Each human-typed line can not be bigger than sizeof(buffer1)-10
//
// 3) Each LISP generated parse can not be bigger than sizeof(buffer1)-10
//
// 4) sizeof(buffer2) must be at least TWICE the size of (buffer1).
//    For reasons, see the comment for the function "local_process_input"
//
// 5) The LISP code should be written, so that whenever it is
//    ready to accept a sentence, it should print out the
//    following 7 characters (not counting the DOUBLE quote): "ready> "
//
// 6) The LISP code should be written, so that whenever
//    it produces a successful parse, it prints out the
//    following 8 characters (not counting the DOUBLE quote): "[OUTPUT]"
//
// 7) This simulator needs to detect whether the LISP has failed or not.
//    Currently, we do this by watching the output from LISP.
//    As soon as we see the string "CL-USER(", we assume LISP has failed.
//    (This is the behavior from the "mlisp" installed in MIT)
//    If your LISP environment generates a different fatal error message,
//    then adjust the code below accordingly.

//=============================================================================

static int local_planeID=1; // Current "plane" id.

char buffer1[16384];   // Must be >= 100 bytes. Preferably much bigger.
char buffer2[16384*2]; // Must be at least twice the size of "buffer1".
static int bufnow=0;     // Number of bytes currently in "buffer1".
static int bufstarted=0; // If 0: LISP isn't initialized yet. Else: it has.

//=============================================================================

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "gui.h"
#include "reparse.h"
#include "dispatch.h"
#include "lisp.h"

//=============================================================================

// Returns TRUE if "xx" is a letter of the alphabet
#define local_is_alpha(xx) (((xx)>='A' && (xx)<='Z')||((xx)>='a' && (xx)<='z'))

// Returns TRUE if "xx" is a numeric digit
#define local_is_digit(xx) ((xx)>='0' && (xx)<='9')

// Returns TRUE if "xx" is a "linebreak", "space", "tab",... etc.
#define local_is_space(xx) ((xx)>0 && (xx)<=32)

// Returns TRUE if the first 7 characters of "line" is "ready> "
static int local_is_ready(char* line)
 {
  if (line[0]!='r') return 0;
  if (line[1]!='e') return 0;
  if (line[2]!='a') return 0;
  if (line[3]!='d') return 0;
  if (line[4]!='y') return 0;
  if (line[5]!='>') return 0; return (line[6]==' ');
 }

// Returns TRUE if the first 8 characters of "line" is "[OUTPUT]"
static int local_is_output(char* line)
 {
  if (line[0]!='[') return 0;
  if (line[1]!='O') return 0;
  if (line[2]!='U') return 0;
  if (line[3]!='T') return 0;
  if (line[4]!='P') return 0;
  if (line[5]!='U') return 0;
  if (line[6]!='T') return 0; return (line[7]==']');
 }

// Returns TRUE if the first 8 characters of "line" is "CL-USER("
static int local_is_CLUSER(char* line)
 {
  if (line[0]!='C') return 0;
  if (line[1]!='L') return 0;
  if (line[2]!='-') return 0;
  if (line[3]!='U') return 0;
  if (line[4]!='S') return 0;
  if (line[5]!='E') return 0;
  if (line[6]!='R') return 0; return (line[7]=='(');
 }

// Returns TRUE if the string "CL-USER(" occurs inside the buffer,
//    (where the buffer currently contains BUFLEN bytes.
static int local_contains_CLUSER(char* buf,int buflen)
 {
  while(buflen>=8)
   {
    if (*buf=='C') if (local_is_CLUSER(buf)) return 1;
    buf=buf+1; buflen=buflen-1;
   }
  return 0;
 }

//=============================================================================

// INPUT:
//   It assumes the user has typed 1 input sentence.
//   And that this sentence is currently stored in "buffer1".
//
// OUTPUT:
//   It cleans up the sentence (eg. insert $ in front of numbers),
//   And stores the revised sentence into "buffer2".
//
// Specifically, it performs the following changes:
//   1) Converts UPPER case to LOWER case.
//   2) Replace all punctuations with spaces.
//   3) Insert $ in front of numbers
//   4) Substitute o'clock with o^clock
//   5) Do very simple-minded expansion for "i'm", "you're", "let's" ...
//   6) Finally: it appends a "newline" character at the end.
//
// Note: At most, this transformation will double the length of string.
// Therefore, sizeof(buffer2) must be at least TWICE the size of "buffer1".

static void local_process_input(void)
 {
  char* src=buffer1; char* dest=buffer2;
  // We need the following line to deal with "leading numbers"
  if (local_is_digit(*src)) { *dest='$'; dest=dest+1; }
  for(;(*src);dest=dest+1,src=src+1)
   {
    // Special processing for "let's"
    if ((*src)=='l' || (*src)=='L')
     if (src[1]=='e' || src[1]=='E')
      if (src[2]=='t' || src[2]=='T')
       if (src[3]=='\'')
        if (src[4]=='s' || src[4]=='S')
         {
          *dest='l'; dest[1]='e'; dest[2]='t'; dest[3]=' '; dest[4]='u';
          dest=dest+4; src=src+3; continue;
         }
    if ((*src)>='a' && (*src)<='z') { *dest=(*src); continue; }
    if ((*src)>='A' && (*src)<='Z') { *dest=((*src)-'A')+'a'; continue; }
    if (local_is_digit(*src)) { *dest=(*src); continue; }
    // The following deals with the SINGLE quote, and other Punctuations
    if ((*src)!='\'')
       { *dest=' '; } // This removes punctuations
    else if (src[1]=='r' && src[2]=='e' && !(local_is_alpha(src[3])))
       { *dest=' '; dest[1]='a'; dest=dest+1; continue; }
    else if (src[1]=='s' && !(local_is_alpha(src[2])))
       { *dest=' '; dest[1]='i'; dest=dest+1; continue; }
    else if (src[1]=='m' && !(local_is_alpha(src[2])))
       { *dest=' '; dest[1]='a'; dest=dest+1; continue; }
    else
       { *dest='^'; }
    // The following line prepends '$' to stand-alone numbers
    if (local_is_digit(src[1])) { dest=dest+1; *dest='$'; }
   }
  // Done. Now terminate the Destination String.
  *dest='\n'; dest[1]=0;
 }

//=============================================================================

// INPUT:
//   It assumes the LISP parser has generated a bunch of output.
//   And that the PARSER output is currently stored in "buffer1".
//
// OUTPUT:
//   It checks to see if the output was a successful parse or not.
//   ==> If successful, it sends the output to our Discourse manager.
//   ==> If failure, it displays the error message on the screen.
//
// Specifically, it performs the following changes:
//   1) It removes all line breaks.
//   2) It replaces adjacent SPACE characters into a single space character.
//   3) It then look for the success indicator.
//      ==> If found, (and there can be multiple success indicators),
//          it send EACH successful parse to our Discourse manager.
//      ==> If not found, it prints the error message on the screen.

static int local_process_output(void)
 {
  char* src=buffer1; char* dest=buffer1; int first=1;
  char answer[1024]; int answerlen=0;

  // First, replace all ASCII control codes by SPACE
  // Then shorten the spaces...
  while(1)
   {
    if ((*src)==0) { *dest=0; break; }
    if (local_is_space(*src))
     {
      while(local_is_space(*src)) src=src+1;
      if ((*src)==0) { *dest=0; break; } *dest=' '; dest=dest+1;
     }
    if ((*src)<0 || (*src)>=127) *src='*'; // Hides non-printable characters
    *dest=(*src); dest=dest+1; src=src+1;
   }

  // Did the parse fail?
  if (!local_is_output(buffer1)) goto bad;

  // Process EACH successful parse (There may be more than one)
  src=buffer1;
  while(1)
   {
    char* scan; scan=src+1;
    // This loop scans for the next "Success Indicator"
    while(1)
     {
      char* peek; if (*scan==0) { scan=NULL; break; } peek=scan;
      while(*peek==' ') peek=peek+1;
      if (*peek=='[') if (local_is_output(peek)) {*scan=0; scan=peek+1; break;}
      if (*peek==0) { scan=NULL; break; } scan=peek+1;
     }
    // The next line "skips" the string "[OUTPUT]"
    while((*src) && (*src!=']')) src=src+1; if (*src) src=src+1;
    // The next line skips any leading spaces
    while(*src==' ') src=src+1;
    // The next 3 lines passes THIS successful parse to our Discourse manager,
    // then accumulate the answer.
    process_line(local_planeID,src,answer+answerlen);
    if (strcasecmp(answer+answerlen,"Roger.")!=0)
     {
      printf("%s\n",answer+answerlen); fflush(stdout);
      gui_text_addstring(answer+answerlen,0); gui_text_addchar('\n',0);
      gui_commit();
      answerlen = answerlen + strlen(answer+answerlen);
     }
    // If there are no more parses, then quit. Else ==> repeat the loop.
    if (scan==NULL) break; src=scan; first=0;
   }

  // If I haven't said any message, then say "Roger."
  if (answerlen==0)
   {
    printf("Roger.\n"); fflush(stdout);
    gui_text_addstring("Roger.\n",0); gui_commit();
   }

  return 1;

  bad:
  printf("I can not understand the sentence. Please repeat.\n"); fflush(stdout);
  gui_text_addstring("I can not understand the sentence. Please repeat.\n",0);
  gui_commit();
  return 0;
 }

//=============================================================================

// Given "file" == the communication pipe
// Given "buf"  == a string of characters that we want to send down the pipe
//
// It tries its best to send the entire string to the pipe.
// ==> Returns 0 if failure
// ==> Returns non-zero if success

static int local_write(int file, char* buf)
 {
  ssize_t result; ssize_t n=0; char* ptr=buf;
  // First compute the length of the string
  while(*ptr) { ptr=ptr+1; n=n+1; if (n<0) return 0; }
  // Send the string to the pipe
  while(n>0)
   {
    if ((result=write(file,buf,n))<=0)
     {
      if (result==0) return 0;
      if (errno==EINTR || errno==EAGAIN || errno==EWOULDBLOCK) continue;
      return 0;
     }
    n=n-result; buf=buf+result;
   }
  return 1;
 }

//=============================================================================

// This function simply does nothing.
// (This function is used as the "default action" for UNIX signals.)

static void local_signal_handler(int unused) { return; }

//=============================================================================

// This function initializes the connection to the LISP parser by...
// (1) Execute the command "mlisp -s air_control.lisp" in the background
// (2) Return the connection "return_in"  by which we can READ from LISP
// (3) Return the connection "return_out" by which we can WRITE to LISP
//
// If failure -> Return 0
// If success -> Return 1 (And "*return_in" and "*return_out" will contain
//                         the pipe that we can use to read/write to LISP)

int lisp_init(int* return_in, int* return_out)
 {
  int error; int in[2],out[2]; pid_t childID;

  // Disable all signals except "SIGINT"
  struct sigaction sig; int i;
  sigemptyset(&(sig.sa_mask));
  sig.sa_flags=(SA_NOCLDSTOP|SA_RESTART);
  sig.sa_handler=local_signal_handler;
  for(i=0; i<100; i=i+1)
   {
    if (i==SIGINT) continue;
    while(sigaction(i,&sig,NULL)!=0)
     {
      if (errno==EINVAL) break; // This means it's not a SETTABLE signal
      if (errno==EAGAIN || errno==EWOULDBLOCK || errno==EINTR) continue;
      printf("ERROR: Can not clear the signals!\n"); return 0;
     }
   }

  // Is the "air_control.lisp" file in the current directory?
  if (access("air_control.lisp",R_OK)!=0)
   {
    printf("ERROR: Can not find \"air_control.lisp\"\n");
    printf("Please run this program in the directory containing that file.\n");
    return 0;
   }

  // Close STDERR
  error=open("/dev/null",O_RDWR,0600);
  if (error<0) return 0;
  if (dup2(error,2)==(-1)) return 0;
  close(error);

  // Create 2 pipes
  if (pipe(in)!=0) return 0;
  if (pipe(out)!=0) return 0;

  // Create a subprocess
  childID=fork(); if (childID==(-1)) return 0;

  // In the subprocess...
  if (childID==0)
   {
    char* args[] = { "mlisp" , "-s", "air_control.lisp" , NULL };
    // The above is the "command line" to run MLISP.
    // If we need to run a different MLISP, change the above line.
    if (dup2(in [0],0)==(-1)) _exit(1);
    if (dup2(out[1],1)==(-1)) _exit(1);
    close(in[1]); close(out[0]); close(in[0]); close(out[1]);
    execvp(args[0],args);
    _exit(1);
   }

  // In the parent process...
  close(in[0]); close(out[1]);

  // Make the "read from LISP" connection a "non-blocking" connection
  fcntl(out[0],F_SETFL,fcntl(out[0],F_GETFL) | O_NONBLOCK);

  // Return success!
  *return_in=out[0];
  *return_out=in[1];
  return 1;
 }

//=============================================================================

// Give "file" is a connection to read from LISP.
// It tries to read a block of data from the LISP program into "buffer1".
// (Note: If LISP hasn't said anything, then it immediately returns 0)
//
// Returns == 0  -> LISP hasn't generated enough output. Please try again soon.
// Returns == 1  -> LISP is newly initialized. And LISP is ready for input.
// Returns == 2  -> LISP is ready for input.
//
// Note: Whenever this function has accumulated enough LISP output,
// it will call "local_process_output",
// which will then send any "successful parse" to the discourse manager,
// and also display any "parser error" to the user.

int lisp_read(int f)
 {
  ssize_t n;
  // Below is a minor optimization: if we haven't started any parsing,
  // and LISP is simply outputting a lot of "WELCOME MESSAGES",
  // then keep the last 20 bytes, and skip the rest.
  // Why? Well, the 20 is long enough to make sure we can still detect
  // the "CL-USER(" prompt, the "ready> " prompt, and the "[OUTPUT]" prompt.
  if (bufstarted==0 && bufnow>20)
     { memmove(buffer1,buffer1+(bufnow-20),20); bufnow=20; }
  // Read messages from LISP
  if ((n=read(f,buffer1+bufnow,sizeof(buffer1)-3-bufnow))<=0)
   {
    if (n==0) { printf("WARNING: LISP has suddenly terminated.\n"); exit(1); }
    if (errno==EINTR || errno==EWOULDBLOCK || errno==EAGAIN) return 0;
    n=errno; printf("WARNING: Failure %d when reading from LISP!",n); exit(1);
   }
  // Append the messages into "buffer1"
  bufnow=bufnow+n;
  // Is there a "CL-USER(" in the message? If so, LISP has failed!
  if (local_contains_CLUSER(buffer1,bufnow))
   {
    if (bufstarted==0) printf(" failed.\n");
    printf("ERROR: The LISP parser has failed.\n"); exit(1);
   }
  // Does the end of the message say "ready> " ???
  if (bufnow>=7) if (local_is_ready(buffer1+(bufnow-7)))
   {
    // Delete the last 7 characters from the buffer
    buffer1[bufnow-7]=0; bufnow=0;
    // If there was a REPLY, then forward the reply to "local_process_output"
    if (!bufstarted)
       { printf(" done. Hit Control-C to exit.\n"); bufstarted=1; return 1; }
    local_process_output();
    return 2;
   }
  // Not enough data yet... Please retry later...
  return 0;
 }

//=============================================================================

// Given a line of "user sentence" in "buffer1".
//
// This function cleans it up a bit, store the new sentence in "buffer2",
// then send the contents of "buffer2" to LISP

void lisp_write(int f,int newplane)
 {
  local_planeID=newplane;
  // Call "local_process_input" to tidy up the user's sentence
  local_process_input();
  // Pass the sentence to the LISP parser. If failure, then EXIT right away
  if (local_write(f,buffer2)==0)
     { printf("WARNING: LISP program has suddenly terminated.\n"); exit(1); }
 }

//=============================================================================
