
//=============================================================================

#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "airport.h"
#include "gui.h"
#include "plane.h"

//=============================================================================

// The value of PI used for doing angle translation
#define pi ((double)(3.1415926535897932384626433832795))

//=============================================================================

static plane** local_allplane=NULL; // The array of planes we've created so far
static int local_planecount=0;      // # of planes created so far...

// Give a plane id, it returns the corresponding plane.
// (If it doesn't exist, it returns NULL)
plane* plane::getplane(int id)
 {
  if (id<=0 || id>local_planecount) return 0;
  return local_allplane[id-1];
 }

//=============================================================================

// Constructor for a new "plane"
plane::plane(int start)
 {
  // Increment "planecount", then enlarge the "allplane" array
  local_planecount++;
  local_allplane
    =(plane**)realloc(local_allplane,sizeof(plane*)*local_planecount);
  if (local_allplane==NULL) exit(1);
  local_allplane[local_planecount-1]=this;
  // Initialize simple fields...
  state=PARK;
  statetime=0;
  altitude_locked=0;
  speed_locked=0;
  id=local_planecount;
  altimeter=0;
  speed=0; targetSpeed=0; oldtargetSpeed=0;
  freq=1; oldfreq=1;
  target_altitude=100;
  newly_changed=NO_CHANGE;
  // Initialize "src", "dest", "x", "y", and "deg"
  src=start;
  dest=start;
  x=airport_points[src].x;
  y=airport_points[src].y;
  deg=0;
  // Initialize "path"
  path=new int[airport_max*3+2];  if (!path) exit(1);
  path[0]=src; path[1]=(-1);
  // Show the plane!
  gui_drawPlane(x,y,0,id);
 }

//=============================================================================

// Given temp = { ? , ? , ? , ... , ? , -1 },
// this chooses it as the new path for the plane.
//
void plane::set_path(int* temp)
 {
  int i,j; int flag;
  if (path[0]>=0 && path[1]>=0 && temp[0]>=0 && temp[1]>=0 && temp[1]!=path[1])
   {
    src=path[1]; path[0]=src;
    // copy temp[0..] to path[1..]
    i=0; while(temp[i]>=0) { path[i+1]=temp[i]; i++; } path[i+1]=(-1); flag=1;
   }
  else
   {
    // copy temp[0..] to path[0..]
    i=0; while(temp[i]>=0) { path[i]=temp[i]; i++; } path[i]=(-1); flag=0;
   }
  // Now, let's optimize the path by removing duplicate points...
  if (path[0]>=0) for(i=flag;path[i]>=0;i++)
   {
    // Is this point repeated later in the path?
    int match=(-1); for(j=i+1;path[j]>=0;j++) if (path[i]==path[j]) match=j;
    // If so, then remove the redundant sub-path
    if (match>=0)
       { for(j=0;;j=j+1) { path[i+j]=path[match+j]; if (path[i+j]<0) break; } }
   }
 }

//=============================================================================

// Put a plane into "landing mode", and commence landing on a particuarl runway
void plane::land(int runway)
 {
  int* newpath = airport_runway_to_points(runway);
  if (newpath==NULL) return;
  path[0]=newpath[0];
  path[1]=newpath[1];
  path[2]=newpath[2];
  airport_optimize(path[2],airport_get_endlot(),path+3);
  x=airport_points[path[0]].x;
  y=airport_points[path[0]].y;
  state=LANDING;
  speed=1;
 }

//=============================================================================

// Given a path = { ? , ? , ? , ... , ? , -1 }, this computes
// the total cost.

static int local_totalcost(int* path)
 {
  int i=0; int cost=0;
  if (path[0]>=0)
   while(path[i+1]>=0)
    {
     int y=path[i+1],x=path[i]; i=i+1; cost=cost+(airport_dist[x][y]);
    }
  return cost;
 }

//=============================================================================

// Choose "newdest" as a new destination, and re-plot a shortest path
void plane::set_dest(int newdest)
 {
  int temp[airport_max+2];
  if (newdest<0 || newdest>=airport_max) return; // No such destination!
  if (airport_optimize(src,newdest,temp)<0) return; // Unreachable!
  dest=newdest;
  set_path(temp);
  newly_changed=LAST_CHANGE_IS_DEST;
  speed=1; state=TAXI; statetime=0;
 }

// Given that we already have a destination,
// this function tries to re-plot a path that includes a taxiway.
void plane::via(int t)
 {
  int i; int temp[airport_max*3+2];
  int minchoice1=(-1),minchoice2=(-1),mincost=0,cost1,cost2,cost3,len;
  int* local_taxiway = airport_taxiway_to_points(t);
  // Sanity check on "dest"...
  if (src==dest || (path[0]>=0 && path[1]==dest)) { set_dest(dest); return; }
  // Is there such a taxiway?
  if (local_taxiway==NULL) { set_dest(dest); return; }
  // For each segment on the taxiway...
  for(i=0; local_taxiway[i]>=0; i=i+1)
   {
    int x=local_taxiway[i],y=local_taxiway[i+1],sum; if (y<0) continue;
    cost3=airport_dist[x][y];
    //
    if (airport_optimize(src,x,temp)<0) goto next; cost1=local_totalcost(temp);
    if (airport_optimize(y,dest,temp)<0)goto next; cost2=local_totalcost(temp);
    sum=cost1+cost2+cost3;
    if (minchoice1<0 || mincost>sum) {minchoice1=x; minchoice2=y; mincost=sum;}
    //
    next:
    if (airport_optimize(src,y,temp)<0) continue; cost1=local_totalcost(temp);
    if (airport_optimize(x,dest,temp)<0)continue; cost2=local_totalcost(temp);
    sum=cost1+cost2+cost3;
    if (minchoice1<0 || mincost>sum) {minchoice1=y; minchoice2=x; mincost=sum;}
   }
  if (minchoice1<0) { set_dest(dest); return; } // No choice what-so-ever!
  // Re-generate the path...
  len=airport_optimize(src,minchoice1,temp);
  airport_optimize(minchoice2,dest,temp+len);
  // Do it!
  set_path(temp);
 }

//=============================================================================

// Given the segment (x0,y0)->(x1,y1)
// Given the plane is currently at (*rx,*ry)
//
// If the airplane hasn't reached (x1,y1) yet
// Then
//   ==> Move it further. Update (*rx) and (*ry) and (*retdeg). Then return 1.
// Else
//   ==> Do nothing. Return 0.

static int local_move(int x0,int y0,int x1,int y1,int* rx,int* ry,int* retdeg)
 {
  int x=(*rx); int y=(*ry); int deg; int step=3;
  if (y==y1)
   {
    if (x==x1) return 0;
    if (x<x1) { x=x+step; if (x>=x1) x=x1; deg=0; }
         else { x=x-step; if (x<=x1) x=x1; deg=180; }
   }
  else if (x==x1)
   {
    if (y<y1) { y=y+step; if (y>=y1) y=y1; deg=90; }
         else { y=y-step; if (y<=y1) y=y1; deg=270; }
   }
  else
   {
    double slope = ((double)(y1-y0)) / ((double)(x1-x0));
    if (x<x1) { x=x+step; if (x>=x1) x=x1; }
        else  { x=x-step; if (x<=x1) x=x1; }
    y=((int)((x-x0)*slope))+(y0);
    if (y0<y1) { if (y<y0) y=y0; if (y>y1) y=y1; }
          else { if (y>y0) y=y0; if (y<y1) y=y1; }
    if ((x1>x0 && y1<y0) || (x1<x0 && y1>y0))
     {
      deg=(int) ((atan(-slope)/(2*pi))*360);
      // Shift the degrees to 0..90
      while(deg<0) deg=deg+360; if (deg>90) deg=90;
      // Mirror it appropriately
      if (x1<x0) { deg=180-deg; } else { deg=(360-deg)%360; }
     }
    else
     {
      deg=(int) ((atan(slope)/(2*pi))*360);
      // Shift the degrees to 0..90
      while(deg<0) deg=deg+360; if (deg>90) deg=90;
      // Rotate it so that it's clockwise...
      if (x1<x0) { deg=deg+180; }
     }
   }
  *retdeg=deg; *rx=x; *ry=y; return 1;
 }

//=============================================================================

// Advance a plane according to what it's being told to do...
int plane::advance(void)
 {
  int dest_runway_choice=(-1),dest_runway=(-1),dest_dist=10000000;
  {
   // Search for my destination runway's "position" location
   if (path[0]>=0 && path[1]>=0 && path[2]<0) dest_runway=path[1];
   if (path[0]>=0 && path[1]<0) dest_runway=path[0];
   if (dest_runway>=0)
    {
     int z;
     for(z=0;z<=35;z++)
      {
       int p = airport_runway_to_point(z); if (p<0 || p!=dest_runway) continue;
       int x0=(airport_points[p].x) - x; int y0=(airport_points[p].y) - y;
       dest_dist=(x0*x0+y0*y0); dest_runway=p; dest_runway_choice=z; break;
      }
    }
  }
  int i,x0,y0,x1,y1,newx,newy,newdeg;
  // Increment the amount of time I have being in my current state
  i=statetime+1; if (i>=0) statetime=i;
  // If I just reached the RUNUP area, then do runup...
  if (state==TAXI && dest_dist<5000)
   {
    state=RUNUP; statetime=0; speed=0; return 0;
   }
  // If I've finished my run-up, then try to get closer...
  if (state==RUNUP && statetime>=60) speed=1;
  // If I've reached the "holding short" area, then start to hold short...
  if (state==RUNUP && dest_dist<1500)
   {
    state=HOLDSHORT; statetime=0; speed=0; return 0;
   }
  // If I've reached the "position", then say "POSITION"
  if (state==HOLDSHORT && dest_dist==0)
   {
    int* list=airport_runway_to_points(dest_runway_choice);
    if (list)
     {
      int x1=airport_points[list[2]].x;
      int y1=airport_points[list[2]].y; int unused1,unused2,unused3;
      local_move(x,y,x1,y1,&unused1,&unused2,&unused3); deg=unused3;
      set_path(list);
      state=POSITION; statetime=0; speed=0; return 0;
     }
   }
  // If I'm stopped, then don't move.
  if (speed<=0) { if (state==TAKINGOFF) set_state(OFFSCREEN); return 0; }
  // Assuming no pausepoint/collisions, try to compute (newx,newy,newdeg)
  while(1)
   {
    if (path[0]<0 || path[1]<0) { speed=0; return 0; }
    newdeg=deg;
    x0=airport_points[path[0]].x; y0=airport_points[path[0]].y; newx=x;
    x1=airport_points[path[1]].x; y1=airport_points[path[1]].y; newy=y;
    if (local_move(x0,y0,x1,y1,&newx,&newy,&newdeg)) break;
    src=path[1]; x0=0; while(path[x0]>=0) { path[x0]=path[x0+1]; x0++; }
   }
  // Am I too close to another plane? If so, don't move forward.
  for(i=1;i<=local_planecount;i++)
   {
    if (i==id || local_allplane[i-1]->state==OFFSCREEN) continue;
    x0=(local_allplane[i-1]->x) - newx; // x0 = horizontal distance
    y0=(local_allplane[i-1]->y) - newy; // y0 = vertical distance
    x0=(x0*x0+y0*y0); // x0 = the square of the raw distance
    if (x0<100) return 0; // Too close. No movement allowed at all.
    if (x0<1000 && id<i); // Somewhat close. But my priority is lower. So stop.
   }
  // No collision detected. So let's actually move the airplane
  x=newx; y=newy; deg=newdeg;
  return 1;
 }

//=============================================================================

// Undraws a plane
void plane::hide(void)
 {
  gui_undrawPlane(x,y);
 }

// Draws a plane (if the plane is not currently off-screen)
void plane::show(void)
 {
  if (state!=OFFSCREEN) gui_drawPlane(x,y,deg,id);
 }

//=============================================================================
