
//=============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "airport.h"
#include "gui.h"
#include "plane.h"
#include "lisp.h"

//=============================================================================

// Returns TRUE if "xx" is a "linebreak", "space", "tab",... etc.
#define local_is_space(xx) ((xx)>0 && (xx)<=32)

//=============================================================================

// The names of the planes
static char* sim_name[]={"Alaska","Boeing","Cessna","Horizon","United",NULL};

// The current plane we're talking to...
static int sim_plane=1;

// The length of the prompt that we've printed on screen.
// For example, if we printed "Alaska> ", then that's 8.
static int prompt_len=0;

// The # of bytes the user already typed (in the current sentence)
static int len=0;

// If 0, we're waiting for LISP output.
// If non zero, we're ready to type more sentences
static int ready=0;

//=============================================================================

// Print a prompt. If the current plane is "Alaska",
// we'll print "Alaska> "
//
static void local_prompt(void)
 {
  gui_text_addchar('\n',0);
  gui_text_addstring(sim_name[sim_plane-1],gui_rgb(20,20,240));
  gui_text_addstring("> ",gui_rgb(20,20,240));
  buffer1[len]=0; gui_text_addstring(buffer1,0);
  prompt_len=strlen(sim_name[sim_plane-1])+2;
 }

// Print a specific message on screen
// ("planeID" indicates which plane issued this message.)
//
static void local_message(int planeID,char* msg)
 {
  int temp; unsigned int color=gui_rgb(0xB0,0x40,0x40);
  if (ready)
   {
    buffer1[len]=0;
    for(temp=len+prompt_len;temp>0;temp--) gui_text_delchar();
   }
  gui_text_addstring(sim_name[planeID-1],color);
  gui_text_addstring(msg,color);
  if (ready)
   {
    local_prompt();
   }
 }

//=============================================================================

// Given planeID=1...
// this function checks to see if this plane should issue any statement.
// If so, it issues a statement.
//
static void local_checkmessage(int planeID)
 {
  plane* ptr=plane::getplane(planeID);
  if (ptr==NULL) return;
  plane_state s=ptr->get_state(); int t=ptr->get_time();
  if (s==PARK && t==0)
   {
    local_message(planeID," to controller: Ready to taxi.\n"); return;
   }
  if (s==RUNUP && t==0)
   {
    local_message(planeID," to controller: Doing runup...\n"); return;
   }
  if (s==HOLDSHORT && t==0)
   {
    local_message(planeID," to controller: Run up completed. Ready for departure.\n"); return;
   }
  if (s==POSITION && t==0)
   {
    local_message(planeID," to controller: Position and hold. Ready for departure.\n"); return;
   }
  if (s==OFFSCREEN && t==60)
   {
    local_message(planeID," to controller: Inbound.\n"); return;
   }
 }

// This functions cycles through all airplanes,
// checking to see if any of them want to issue a statement.
static void local_checkallmessage(void)
 {
  int i=1;
  while(1) { if (plane::getplane(i)==NULL) break; local_checkmessage(i); i++; }
 }

//=============================================================================

// This function is called when a user has hit ENTER...
// This function will check which plane the user is talking to,
// then send the sentence to the LISP parser.
static void local_sendsentence(int return_out)
 {
  char* ptr; int i; int oldlen;
  //
  // Print a "NEWLINE" if it will make the screen display clearer
  //
  if ((len+7)%gui_textwidth!=0) { gui_text_addchar('\n',0); gui_commit(); }
  buffer1[len]='\n'; buffer1[len+1]=0;
  printf("%s",buffer1); fflush(stdout);
  //
  // If the user has typed a blank line, then do nothing.
  //
  for(ptr=buffer1;local_is_space(*ptr);ptr=ptr+1);
  if (*ptr==0)
   {
    len=0; printf("\n%s> ",sim_name[sim_plane-1]); fflush(stdout);
    local_prompt(); gui_commit(); return;
   }
  //
  // If there is a "airplane designation" in front of it, then get it out
  //
  for(oldlen=strlen(buffer1),i=0;sim_name[i]!=NULL;i++)
   {
    int n=strlen(sim_name[i]);
    if (buffer1[n]!=',') continue;
    buffer1[n]=0;
    if (strcasecmp(buffer1,sim_name[i])!=0) { buffer1[n]=','; continue; }
    memmove(buffer1,buffer1+n+1,oldlen-(n+1)+1); sim_plane=(i+1);
   }
  //
  // Send the setence to LISP.
  //
  lisp_write(return_out,sim_plane); ready=0;
 }

//=============================================================================

// This function initializes the simulation environment.
//
// Right now, we create Alaska+Boeing+Cessna all in the starting
// parking lot, all wanting to take off (eventually).
//
// And we have Horizon+United, both of whom (in about 30 time units)
// will want to land.

void siminit(void)
 {
  int start=airport_get_startlot(), end=airport_get_endlot(), i;
  gui_OpenWindow("aircontrol\0\0","Air Control Simulator\0\0",0,0);
  //
  plane* p1=new plane(start);p1->set_dest(end); for(i=0;i<12;i++)p1->advance();
  plane* p2=new plane(start);p2->set_dest(end); for(i=0;i<8;i++)p2->advance();
  plane* p3=new plane(start);p3->set_dest(end); p3->advance();
  plane* p4=new plane(start);p4->set_dest(end);
  plane* p5=new plane(start);p5->set_dest(end);
  //
  // In the above call, I call "advance()" a few times in order
  // to space them out a little bit.
  //
  p1->set_state(PARK     ); p1->speed=0;
  p2->set_state(PARK     ); p2->speed=0;
  p3->set_state(PARK     ); p3->speed=0;
  p4->set_state(OFFSCREEN); p4->speed=0;
  p5->set_state(OFFSCREEN); p5->speed=0;
 }

void animate(int advance)
 {
  int i; plane* ptr;

  // Check for new messages from each plane...
  local_checkallmessage();

  // Undraw each plane
  for(i=1;;i++) { ptr=plane::getplane(i); if (!ptr) break; ptr->hide(); }

  // Advance each plane
  for(i=1;;i++) { ptr=plane::getplane(i); if (!ptr) break; ptr->advance(); }

  // Draw each plane again
  for(i=1;;i++) { ptr=plane::getplane(i); if (!ptr) break; ptr->show(); }

  // Commit the graphical change to the screen.
  gui_commit();
 }

//=============================================================================

int main()
 {
  int return_in,return_out,key,GUIready=0;
  printf("\nSystem initializing..."); fflush(stdout);
  lisp_init(&return_in,&return_out);
  while(1)
   {
    key=gui_HandleEvents(return_in);
    if (key==0 && GUIready) animate(10); // If nothing to do, then animate.
    if (key==27) break;                  // If user hits ESC, then exit.
    if (ready==0) // If we're waiting for LISP...
     {
      ready=lisp_read(return_in);
      if (ready==0) continue;
      if (ready==1) { siminit(); GUIready=1; }
      len=0; printf("\n%s> ",sim_name[sim_plane-1]); fflush(stdout);
      local_prompt(); gui_commit(); continue;
     }
    if (key>=32 && key<=126)
     {
      // User hit a regular key. So add this key to the sentence.
      buffer1[len]=key; len++; gui_text_addchar(key,0); gui_commit();
     }
    if (key==8 && len>0)
     {
      // User hit a BACKSPACE. So if sentence_len>0, remove the last character.
      gui_text_delchar(); gui_commit(); len--;
     }
    if (key==10 || key==13) local_sendsentence(return_out); // User hits ENTER
   }
  gui_CloseWindow();
  printf("\nSystem shutting down... done.\n"); fflush(stdout);
  return 0;
 }

//=============================================================================
